import aframexr
import unittest

from aframexr.utils.constants import AVAILABLE_ENVIRONMENTS, ERROR_MESSAGES
from .constants import *


class TestConeOK(unittest.TestCase):
    """Tests for simple cone creation."""

    def test_simple(self):
        """Simple cone creation."""
        cone = aframexr.Cone()
        cone.to_html()

    def test_from_dict(self):
        """Cone creation using form_dict() method."""
        cone = aframexr.Cone.from_dict(aframexr.Cone().to_dict())
        cone.to_html()

    def test_from_json(self):
        """Cone creation using form_json() method."""
        cone = aframexr.Cone.from_json(aframexr.Cone().to_json())
        cone.to_html()

    def test_color(self):
        """Cone with specific color creation."""
        for c in SIMPLE_ELEMENTS_COLORS:
            cone = aframexr.Cone(color=c)
            cone.to_html()

    def test_height(self):
        """Cone with specific height creation."""
        for h in ALL_MARK_DEPTHS_HEIGHTS_WIDTHS:
            cone = aframexr.Cone(height=h)
            cone.to_html()

    def test_position(self):
        """Cone with specific position creation."""
        for p in POSITIONS + POSITION_FORMATS:
            cone = aframexr.Cone(position=p)
            cone.to_html()

    def test_radius_bottom(self):
        """Cone with specific bottom's radius creation."""
        for r in RADIUS:
            cone = aframexr.Cone(radius_bottom=r)
            cone.to_html()

    def test_radius_top(self):
        """Cone with specific top's radius creation."""
        for r in RADIUS:
            cone = aframexr.Cone(radius_top=r)
            cone.to_html()

    def test_all_parameters(self):
        """Cone creation with all parameters specified."""
        for c, h, p, rb, rt in zip(SIMPLE_ELEMENTS_COLORS, ALL_MARK_DEPTHS_HEIGHTS_WIDTHS, POSITIONS, RADIUS,
                                   RADIUS):
            cone = aframexr.Cone(color=c, height=h, position=p, radius_bottom=rb, radius_top=rt)
            cone.to_html()

    def test_concatenation(self):
        """Concatenation of simple cones."""
        concatenated_chart = aframexr.Cone(position=CONCATENATION_POSITIONS[0])
        for p in CONCATENATION_POSITIONS[1:]:
            concatenated_chart += aframexr.Cone(position=p)
        concatenated_chart.to_html()

    def test_environment(self):
        """Scene creation with personalized environment."""
        for e in AVAILABLE_ENVIRONMENTS:
            aframexr.Cone().to_html(environment=e)

    def test_movable(self):
        """Movable cone creation."""
        html = aframexr.Cone().movable().to_html()
        self.assertIn('movable', html)


class TestConeERROR(unittest.TestCase):
    """Tests for cone creation errors."""

    def test_color_error(self):
        """Cone with specific color creation error."""
        not_an_string = ['not a color']
        with self.assertRaises(TypeError) as error:
            # noinspection PyTypeChecker
            cone = aframexr.Cone(color=not_an_string)
            cone.to_html()
        self.assertEqual(
            str(error.exception),
            ERROR_MESSAGES['TYPE'].format(
                param_name='specs.color', expected_type=str.__name__, current_type=type(not_an_string).__name__
            )
        )

    def test_radius_bottom_error(self):
        """Cone with specific bottom's radius creation error."""
        for r in NOT_GREATER_THAN_0_RADIUS:
            with self.assertRaises(ValueError) as error:
                cone = aframexr.Cone(radius_bottom=r)
                cone.to_html()
            self.assertEqual(
                str(error.exception),
                ERROR_MESSAGES['POSITIVE_NUMBER'].format(param_name='specs.radius_bottom')
            )

    def test_radius_top_error(self):
        """Cone with specific top's radius creation error."""
        for r in NOT_GREATER_THAN_0_RADIUS:
            with self.assertRaises(ValueError) as error:
                cone = aframexr.Cone(radius_top=r)
                cone.to_html()
            self.assertEqual(
                str(error.exception),
                ERROR_MESSAGES['POSITIVE_NUMBER'].format(param_name='specs.radius_top')
            )

    def test_environment_error(self):
        """Scene creation with invalid personalized environment."""
        environment = 'bad_environment'
        cone = aframexr.Cone()
        with self.assertRaises(ValueError) as error:
            # noinspection PyTypeChecker
            cone.to_html(environment=environment)
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENVIRONMENT'].format(environment=environment))



"""AframeXR entities HTML creator"""

import io
import json
import os
import polars as pl
import urllib.error, urllib.request
import warnings

from functools import lru_cache
from polars import DataFrame

from .axis_creator import AxisCreator
from .chart_creator import ChartCreator
from .constants import ENTITY_IS_MOVABLE, LABELS_SCALE
from .element_creator import ElementCreator, TextCreator


@lru_cache  # Use come cache for increasing performance
def _get_data_from_url(url: str) -> DataFrame:
    """Loads the data from the URL (could be a local path) and returns it as a DataFrame."""
    if url.startswith(('http://', 'https://')):  # Data is stored in a URL
        try:
            with urllib.request.urlopen(url, timeout=10) as response:
                file_type = response.info().get_content_type()
                data = io.BytesIO(response.read())  # For polars
        except urllib.error.URLError:
            raise IOError(f'Could not load data from URL: {url}.')
    else:  # Data is stored in a local file
        path = os.path.normpath(url)
        if not os.path.exists(path):
            raise FileNotFoundError(f'Local file "{path}" was not found.')

        data = open(path, 'rb')
        _, file_type = os.path.splitext(path)
        file_type = file_type.lower()

    try:
        if 'csv' in file_type:  # Data is in CSV format
            df_data = pl.read_csv(data)
        elif 'json' in file_type:
            json_data = json.load(data)
            df_data = DataFrame(json_data)
        else:
            raise ValueError(f'Unsupported file type: {file_type}.')

        return df_data
    except ValueError:
        raise  # To raise previous ValueError
    except Exception as e:
        raise IOError(f'Error when processing data. Error: {e}.')

    finally:
        if data and not url.startswith(('http', 'https')):
            data.close()  # Close the file


def _get_raw_data_and_params(chart_specs: dict) -> tuple[DataFrame, set]:
    """
    Returns a tuple containing the raw data from the chart specifications (transformed if necessary),
    and a set containing the names for the params of the chart.
    """
    # Get the raw data of the chart
    data_field = chart_specs['data']
    if data_field.get('url'):  # Data is stored in a file
        raw_data = _get_data_from_url(data_field['url'])
    elif data_field.get('values'):  # Data is stored as the raw data
        json_data = data_field['values']
        raw_data = DataFrame(json_data)
    else:  # pragma: no cover (should never enter here, as chart_specs should have previously been validated)
        raise RuntimeError('Unreachable code: chart_specs should have been validated earlier')

    # Transform data (if necessary)
    from ..api.aggregate import AggregatedFieldDef  # To avoid circular import error
    from ..api.filters import FilterTransform

    transform_field = chart_specs.get('transform')
    params_names = set()
    if transform_field:

        for filter_transformation in transform_field:  # The first transformations are the filters
            if filter_transformation.get('filter'):
                filter_specs = filter_transformation['filter']
                if 'param' in filter_specs:
                    params_names.add(filter_specs['param'])
                else:  # Exclude params from filters
                    if isinstance(filter_specs, str):
                        filter_object = FilterTransform.from_equation(filter_specs)
                    elif isinstance(filter_specs, dict):
                        filter_object = FilterTransform.from_dict(filter_transformation['filter'])
                    else:  # pragma: no cover (should never enter here, as filter specs should have been validated)
                        raise RuntimeError('Unreachable code. Filter specifications should have been validated earlier')

                    raw_data = filter_object.get_filtered_data(raw_data)
                    if raw_data.is_empty():  # Data does not contain any value for the filter
                        warnings.warn(f'Data does not contain values for the filter: {filter_transformation["filter"]}')

        for non_filter_transf in transform_field:  # Non-filter transformations
            groupby = set(non_filter_transf.get('groupby')) if non_filter_transf.get('groupby') else set()
            if non_filter_transf.get('aggregate'):
                for aggregate in non_filter_transf.get('aggregate'):
                    aggregate_object = AggregatedFieldDef.from_dict(aggregate)

                    encoding_channels = {  # Using a set to have the possibility of getting differences
                        ch_spec['field'] for ch_spec in chart_specs['encoding'].values()  # Take the encoding channels
                        if ch_spec['field'] != aggregate_object.as_field  # Except the aggregate field channel
                    }

                    if groupby:
                        not_defined_channels = encoding_channels - set(groupby)  # Difference between sets
                        if not_defined_channels:  # There are channels in encoding_channels not defined in groupby
                            raise ValueError(
                                f'Encoding channel(s) "{not_defined_channels}" must be defined in aggregate groupby: '
                                f'{groupby}, otherwise that fields will disappear.'
                            )
                    else:
                        groupby = encoding_channels  # Use the encoding channels as groupby
                    raw_data = aggregate_object.get_aggregated_data(raw_data, list(groupby))

    # Aggregate in encoding
    encoding_channels_values = list(chart_specs['encoding'].values())
    groupby_fields = [ch['field'] for ch in encoding_channels_values if not ch.get('aggregate')]

    for ch in encoding_channels_values:
        aggregate_op = ch.get('aggregate')
        if aggregate_op is not None:
            aggregate_object = AggregatedFieldDef(aggregate_op, ch['field'])
            raw_data = aggregate_object.get_aggregated_data(raw_data, groupby_fields)

    return raw_data, params_names


def _get_param_combinations(data: DataFrame, param_specs: dict | None) -> list[dict]:
    """Returns a list containing the combinations in data for param specifications."""
    combinations = []

    if param_specs is None:
        return combinations

    select_type = param_specs['select']['type']
    if select_type == 'point':
        combinations = data.select(param_specs['select']['fields']).unique().to_dicts()

    return combinations


class ChartsHTMLCreator:
    """Charts HTML creator class."""

    @staticmethod
    def _create_chart_html(chart_object: ChartCreator, param_name: str = None, param_values: dict = None) -> str:
        filtered_by_params = False
        attributes = ''
        if param_name is not None and param_values is not None:
            filtered_by_params = True
            param_values_str = '__'.join(f'{value}' for value in param_values.values())
            attributes += f' param-name="{param_name}__{param_values_str}" visible="false"'

        chart_html = (f'\t\t\t<a-entity position="{chart_object.get_relative_bottom_left_corner_position()}"'
                      f'{attributes}>\n')
        elements = chart_object.get_elements(filtered_by_params=filtered_by_params)  # Specifications for each element
        for element in elements:
            chart_html += '\t\t\t\t' + element.get_element_html() + '\n'  # Tabulate the lines (better visualization)

        # Axis HTML
        axes_specs = chart_object.get_axes_specs()

        for ax, ax_specs in axes_specs.items():
            chart_html += f'\n\t\t\t\t<!-- {ax.upper()}-axis -->\n'  # Added HTML comment for better visualization
            chart_html += '\t\t\t\t' + AxisCreator.create_axis_html(ax_specs['start'], ax_specs['end']) + '\n'
            for label_pos, label_value in zip(ax_specs['labels_pos'], ax_specs['labels_values']):
                chart_html += '\t\t\t\t' + TextCreator({
                    'value': label_value, 'position': label_pos, 'rotation': ax_specs['labels_rotation'],
                    'align': ax_specs['labels_align'], 'scale': LABELS_SCALE
                }).get_element_html() + '\n'

        # Title
        title_elements = chart_object.get_title_elements(filtered_by_params=filtered_by_params)
        if title_elements:
            chart_html += f'\n\t\t\t\t<!-- Title -->\n'  # Added HTML comment for better visualization
            for element in title_elements:
                chart_html += '\t\t\t\t' + element.get_element_html() + '\n'

        # Legend
        legend_elements = chart_object.get_legend_elements(filtered_by_params=filtered_by_params)
        if legend_elements:
            chart_html += f'\n\t\t\t<!-- Legend -->\n'  # Added HTML comment for better visualization
            for element in legend_elements:
                chart_html += '\t\t\t' + element.get_element_html() + '\n'

        # Close the groups
        chart_html += '\t\t\t</a-entity>\n'
        return chart_html

    @staticmethod
    def _create_element_html(element_specs: dict, is_movable: bool = ENTITY_IS_MOVABLE) -> str:
        element_type = element_specs['element']
        element_object = ElementCreator.create_object(element_type, element_specs)
        return element_object.get_element_html(is_movable=is_movable)

    @staticmethod
    def _create_entity_html(chart_specs: dict, scene_params_map: dict) -> str:
        """
        Returns the HTML of the elements that compose the entity.

        Parameters
        ----------
        chart_specs : dict
            Chart specifications.
        scene_params_map : dict
            Parameters of the scene.

        Notes
        -----
        Supposing that chart_specs is a dictionary (at this method has been called from self.create_charts_html).

        Suppose that the parameters are correct for method calls of ChartCreator and AxisCreator.
        """
        is_movable = chart_specs.get('movable', ENTITY_IS_MOVABLE)

        if 'mark' in chart_specs:  # Chart
            chart_type = chart_specs['mark']['type'] if isinstance(chart_specs['mark'], dict) else chart_specs['mark']
            raw_data, chart_params_names = _get_raw_data_and_params(chart_specs)
            chart_specs['data'] = {'values': raw_data.to_dicts()}
            chart_object = ChartCreator.create_object(chart_type, chart_specs)  # Create the chart object
            group_specs = chart_object.get_group_specs()  # Get the base specifications of the group of elements

            attributes = ''.join(f' {key.replace("_", "-")}="{value}"' for key, value in group_specs.items())
            if is_movable:
                attributes += ' movable'  # For drag-controls

            html = ('<a-entity{attributes}>'.format(attributes=attributes) +
                          "  <!-- Chart's box (modify this values if you want to change position or rotation) -->\n")

            # =================
            if chart_params_names:  # Chart is filtered using params
                for param_name in chart_params_names:
                    param_specs = scene_params_map.get(param_name)

                    if param_specs is None:
                        warnings.warn(
                            f'Parameter {param_name} not found in scene\'s specifications, charts transformed by '
                            f'that param will not be displayed. Make sure the name is correct'
                        )

                    # Create one chart per combination
                    charts_html_list = []
                    param_combinations = _get_param_combinations(raw_data, param_specs)
                    if not param_combinations:
                        new_chart_specs = {**chart_specs, 'data': {'values': raw_data.to_dicts()}}
                        new_chart_object = ChartCreator.create_object(chart_type, new_chart_specs)
                        charts_html_list.append(ChartsHTMLCreator._create_chart_html(new_chart_object))
                    else:
                        for combination in param_combinations:
                            new_data = raw_data
                            for key, value in combination.items():
                                new_data = new_data.filter(pl.col(key) == value)

                            new_chart_specs = {**chart_specs, 'data': {'values': new_data.to_dicts()}}
                            new_chart_object = ChartCreator.create_object(chart_type, new_chart_specs)

                            charts_html_list.append(
                                ChartsHTMLCreator._create_chart_html(
                                    new_chart_object,
                                    param_name=param_name, param_values=combination
                                )
                            )

                    html += '\n'.join(charts_html_list)

            else:
                chart_specs['data'] = {'values': raw_data.to_dicts()}
                html += ChartsHTMLCreator._create_chart_html(chart_object)

            # Close the entity
            html += '\t\t</a-entity>\n\t'

        elif 'element' in chart_specs:  # Single element
            html = ChartsHTMLCreator._create_element_html(chart_specs, is_movable=is_movable)

        else:  # pragma: no cover (should never enter here, as chart_specs should have previously been validated)
            raise RuntimeError('Unreachable code: chart_specs should have been validated earlier')

        return html

    @staticmethod
    def create_charts_html(specs: dict) -> str:
        """
        Returns the HTML of the charts that compose the scene.

        Parameters
        ----------
        specs : dict
            Specifications of all the charts composing the scene.

        Notes
        -----
        Supposing that specs is a dictionary, at this method has been called from SceneCreator.create_scene().

        Suppose that chart_specs is a dictionary for self._create_entity_html(chart_specs).
        """
        scene_params = list(specs.get('params', []))
        for chart in specs.get('concat', []):
            scene_params.extend(chart.get('params', []))
        scene_params_map = {p['name']: p for p in scene_params}

        charts_list = specs.get('concat')
        if charts_list:
            return '\n\t\t'.join(
                ChartsHTMLCreator._create_entity_html(chart, scene_params_map) for chart in charts_list
            )

        return ChartsHTMLCreator._create_entity_html(specs, scene_params_map)



import aframexr
import unittest

from aframexr.utils.constants import AVAILABLE_ENVIRONMENTS, ERROR_MESSAGES
from .constants import *


class TestBoxOK(unittest.TestCase):
    """Tests for simple box creation."""

    def test_simple(self):
        """Simple box creation."""
        box = aframexr.Box()
        box.to_html()

    def test_from_dict(self):
        """Box creation using form_dict() method."""
        box = aframexr.Box.from_dict(aframexr.Box().to_dict())
        box.to_html()

    def test_from_json(self):
        """Box creation using form_json() method."""
        box = aframexr.Box.from_json(aframexr.Box().to_json())
        box.to_html()

    def test_color(self):
        """Box with specific color creation."""
        for c in SIMPLE_ELEMENTS_COLORS:
            box = aframexr.Box(color=c)
            box.to_html()

    def test_depth(self):
        """Box with specific depth creation."""
        for d in ALL_MARK_DEPTHS_HEIGHTS_WIDTHS:
            box = aframexr.Box(depth=d)
            box.to_html()

    def test_height(self):
        """Box with specific height creation."""
        for h in ALL_MARK_DEPTHS_HEIGHTS_WIDTHS:
            box = aframexr.Box(height=h)
            box.to_html()

    def test_position(self):
        """Box with specific position creation."""
        for p in POSITIONS + POSITION_FORMATS:
            box = aframexr.Box(position=p)
            box.to_html()

    def test_rotation(self):
        """Box with specific rotation creation."""
        for r in ROTATIONS + ROTATION_FORMATS:
            box = aframexr.Box(rotation=r)
            box.to_html()

    def test_width(self):
        """Box with specific width creation."""
        for w in ALL_MARK_DEPTHS_HEIGHTS_WIDTHS:
            box = aframexr.Box(width=w)
            box.to_html()

    def test_all_parameters(self):
        """Box creation with all parameters specified."""
        for c, d, h, p, r, w in zip(SIMPLE_ELEMENTS_COLORS, ALL_MARK_DEPTHS_HEIGHTS_WIDTHS,
                                    ALL_MARK_DEPTHS_HEIGHTS_WIDTHS, POSITIONS, ROTATIONS,
                                    ALL_MARK_DEPTHS_HEIGHTS_WIDTHS):
            box = aframexr.Box(color=c, depth=d, height=h, position=p, rotation=r, width=w)
            box.to_html()

    def test_concatenation(self):
        """Concatenation of simple boxes."""
        concatenated_chart = aframexr.Box(position=CONCATENATION_POSITIONS[0])
        for p in CONCATENATION_POSITIONS[1:]:
            concatenated_chart += aframexr.Box(position=p)
        concatenated_chart.to_html()

    def test_environment(self):
        """Scene creation with personalized environment."""
        for e in AVAILABLE_ENVIRONMENTS:
            aframexr.Box().to_html(environment=e)

    def test_movable(self):
        """Movable box creation."""
        html = aframexr.Box().movable().to_html()
        self.assertIn('movable', html)


class TestBoxERROR(unittest.TestCase):
    """Tests for box creation errors."""

    def test_color_error(self):
        """Box with specific color creation error."""
        not_an_string = ['not a color']
        with self.assertRaises(TypeError) as error:
            # noinspection PyTypeChecker
            box = aframexr.Box(color=not_an_string)
            box.to_html()
        self.assertEqual(
            str(error.exception),
            ERROR_MESSAGES['TYPE'].format(
                param_name='specs.color', expected_type=str.__name__, current_type=type(not_an_string).__name__
            )
        )

    def test_environment_error(self):
        """Scene creation with invalid personalized environment."""
        environment = 'bad_environment'
        box = aframexr.Box()
        with self.assertRaises(ValueError) as error:
            # noinspection PyTypeChecker
            box.to_html(environment=environment)
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENVIRONMENT'].format(environment=environment))



import aframexr
import math
import unittest

from bs4 import BeautifulSoup

from aframexr.utils.constants import AVAILABLE_ENVIRONMENTS, ERROR_MESSAGES
from aframexr.api.filters import FilterTransform
from tests.constants import *  # Constants used for testing


def _all_theta_sum_is_360_degrees(pie_chart_html: str) -> bool:
    """Verify that the sum for the theta length of every slice is 360 degrees."""
    total_theta_length = 0

    soup = BeautifulSoup(pie_chart_html, 'lxml')
    slices = soup.find_all('a-cylinder')
    for s in slices:
        total_theta_length += float(s['theta-length'])
    if not math.isclose(total_theta_length, 360):
        print(f'\nDEBUG: total theta length must be 360 degrees.\n\t- Total theta length: {total_theta_length}.')
        return False
    return True

def _slices_are_well_placed(pie_chart_html: str) -> bool:
    """Verify that the slices are well-placed in the pie chart (relative position has to be the same for all)."""
    soup = BeautifulSoup(pie_chart_html, 'lxml')
    slices = soup.find_all('a-cylinder')
    for s in slices:
        if s['position'] != '0 0 0':
            print(f'\nDEBUG: one slice does not have position "0 0 0".\n\t- Position: {s['position']}.')
            return False
    return True


class TestMarkArcOK(unittest.TestCase):
    """Pie chart OK tests."""
    def test_simple(self):
        """Simple pie chart creation."""
        pie_chart = aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales')
        pie_chart_html = pie_chart.to_html()
        self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
        self.assertTrue(_slices_are_well_placed(pie_chart_html))

    def test_simple_with_pandas_not_installed(self):
        """Simple pie chart creation without having pandas installed."""
        import sys
        import importlib
        from unittest import mock

        sys.modules.pop('aframexr', None)
        sys.modules.pop('aframexr.api.components', None)

        with mock.patch.dict(sys.modules, {'pandas': None}):
            import aframexr
            importlib.reload(aframexr)

            pie_chart = aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales')
            pie_chart_html = pie_chart.to_html()
            self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
            self.assertTrue(_slices_are_well_placed(pie_chart_html))

            import aframexr.api.components as components
            self.assertIsNone(components.pd)
            self.assertIs(components.DataFrame, object)

    def test_from_json(self):
        """Pie chart using from_json() method creation."""
        json_string = aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales').to_json()
        pie_chart = aframexr.Chart.from_json(json_string)
        pie_chart_html = pie_chart.to_html()
        self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
        self.assertTrue(_slices_are_well_placed(pie_chart_html))
        self.assertEqual(pie_chart.to_json(), json_string)

    def test_data_format(self):
        """Pie chart changing data format creation."""
        for d in DATA_FORMATS:
            pie_chart = aframexr.Chart(d).mark_arc().encode(color='model', theta='sales')
            pie_chart_html = pie_chart.to_html()
            self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
            self.assertTrue(_slices_are_well_placed(pie_chart_html))

    def test_depth(self):
        """Pie chart changing depth creation."""
        for d in ALL_MARK_DEPTHS_HEIGHTS_WIDTHS:
            pie_chart = aframexr.Chart(DATA, depth=d).mark_arc().encode(color='model', theta='sales')
            pie_chart_html = pie_chart.to_html()
            self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
            self.assertTrue(_slices_are_well_placed(pie_chart_html))

    def test_position(self):
        """Pie chart changing position creation."""
        for p in POSITIONS:
            pie_chart = aframexr.Chart(DATA, position=p).mark_arc().encode(color='model', theta='sales')
            pie_chart_html = pie_chart.to_html()
            self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
            self.assertTrue(_slices_are_well_placed(pie_chart_html))

    def test_position_format(self):
        """Pie chart changing position format creation."""
        for p in POSITION_FORMATS:
            pie_chart = aframexr.Chart(DATA, position=p).mark_arc().encode(color='model', theta='sales')
            pie_chart_html = pie_chart.to_html()
            self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
            self.assertTrue(_slices_are_well_placed(pie_chart_html))

    def test_rotation(self):
        """Pie chart changing rotation creation."""
        for r in ROTATIONS:
            pie_chart = aframexr.Chart(DATA, rotation=r).mark_arc().encode(color='model', theta='sales')
            pie_chart_html = pie_chart.to_html()
            self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
            self.assertTrue(_slices_are_well_placed(pie_chart_html))

    def test_rotation_format(self):
        """Pie chart changing rotation format creation."""
        for r in ROTATION_FORMATS:
            pie_chart = aframexr.Chart(DATA, rotation=r).mark_arc().encode(color='model', theta='sales')
            pie_chart_html = pie_chart.to_html()
            self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
            self.assertTrue(_slices_are_well_placed(pie_chart_html))

    def test_title(self):
        """Pie chart with title creation."""
        pie_chart = aframexr.Chart(DATA, title='Title').mark_arc().encode(color='model', theta='sales')
        pie_chart_html = pie_chart.to_html()
        self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
        self.assertTrue(_slices_are_well_placed(pie_chart_html))

    def test_radius(self):
        """Pie chart changing radius creation."""
        for r in RADIUS:
            pie_chart = aframexr.Chart(DATA).mark_arc(radius=r).encode(color='model', theta='sales')
            pie_chart_html = pie_chart.to_html()
            self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
            self.assertTrue(_slices_are_well_placed(pie_chart_html))

    def test_filter(self):
        """Pie chart changing filter creation."""
        for eq in FILTER_EQUATIONS:
            for f in [eq, FilterTransform.from_equation(eq)]:  # Filter using equation and using FilterTransform object
                pie_chart = aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales').transform_filter(f)
                pie_chart_html = pie_chart.to_html()
                self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
                self.assertTrue(_slices_are_well_placed(pie_chart_html))

    def test_dynamic_filter(self):
        """Pie chart with dynamic filter creation."""
        pie_chart_1 = aframexr.Chart(DATA).mark_arc().encode(color='motor', theta='sales').add_params(DYNAMIC_FILTER)
        pie_chart_2 = (aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales')
                       .transform_filter(DYNAMIC_FILTER))
        final_chart = pie_chart_1 + pie_chart_2
        final_chart_html = final_chart.to_html()
        self.assertTrue(_slices_are_well_placed(final_chart_html))


    def test_several_filters(self):
        """Pie chart with several filters' creation."""
        pie_chart = (aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales')
                     .transform_filter(SEVERAL_FILTER_EQUATIONS[0]))
        for eq in SEVERAL_FILTER_EQUATIONS[1:]:
            pie_chart.transform_filter(eq)

        pie_chart_html = pie_chart.to_html()
        self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
        self.assertTrue(_slices_are_well_placed(pie_chart_html))

    def test_aggregate(self):
        """Pie chart changing aggregates creation."""
        for a in AGGREGATES:
            pie_chart = (aframexr.Chart(DATA).mark_arc().encode(color='model', theta='new_field')
                         .transform_aggregate(new_field=f'{a}(sales)'))
            pie_chart_html = pie_chart.to_html()
            self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
            self.assertTrue(_slices_are_well_placed(pie_chart_html))

            pie_chart_2 = aframexr.Chart(DATA).mark_arc().encode(color='model', theta=f'{a}(sales)')
            pie_chart_2_html = pie_chart_2.to_html()
            self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_2_html))
            self.assertTrue(_slices_are_well_placed(pie_chart_2_html))

    def test_aggregate_position_rotation_radius_filter(self):
        """Pie chart changing position, rotation, radius and filter creation."""
        for agg, pos, rot, rad, fil in zip(AGGREGATES, POSITIONS, ROTATIONS, RADIUS, FILTER_EQUATIONS):
            pie_chart = (aframexr.Chart(DATA, position=pos, rotation=rot).mark_arc(radius=rad)
                         .encode(color='model',theta='agg').transform_filter(fil)
                         .transform_aggregate(agg=f'{agg}(sales)'))
            pie_chart_html = pie_chart.to_html()
            self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
            self.assertTrue(_slices_are_well_placed(pie_chart_html))

    def test_concatenation(self):
        """Pie chart concatenation creation."""
        concatenated_chart = (aframexr.Chart(DATA, position=CONCATENATION_POSITIONS[0]).mark_arc()
                              .encode(color='model', theta='sales'))
        for pos in CONCATENATION_POSITIONS[1:]:
            concatenated_chart += aframexr.Chart(DATA, position=pos).mark_arc().encode(color='model', theta='sales')

        concatenated_chart_html = concatenated_chart.to_html()
        self.assertTrue(_slices_are_well_placed(concatenated_chart_html))

    def test_environment(self):
        """Scene creation with personalized environment."""
        for e in AVAILABLE_ENVIRONMENTS:
            pie_chart = aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales')
            pie_chart_html = pie_chart.to_html(environment=e)
            self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
            self.assertTrue(_slices_are_well_placed(pie_chart_html))

    def test_save(self):
        """Pie chart saving."""
        import tempfile
        from pathlib import Path

        with tempfile.TemporaryDirectory() as tmpdir:
            temp_html_file_path = Path(tmpdir) / "test.html"
            temp_json_file_path = Path(tmpdir) / "test.json"

            pie_chart = aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales')
            pie_chart.save(str(temp_html_file_path))
            pie_chart.save(str(temp_json_file_path))

            self.assertTrue(temp_html_file_path.exists())
            self.assertTrue(temp_json_file_path.exists())

    def test_properties(self):
        """Pie chart properties definition."""
        pie_chart = aframexr.Chart().mark_arc().encode(color='model', theta='sales')
        for p, r in zip(POSITIONS, ROTATIONS):
            pie_chart_2 = pie_chart.properties(data=DATA, position=p, rotation=r)
            pie_chart_2_html = pie_chart_2.to_html()
            self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_2_html))
            self.assertTrue(_slices_are_well_placed(pie_chart_2_html))

    def test_movable(self):
        """Movable pie chart creation."""
        pie_chart_html = aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales').movable().to_html()
        self.assertIn('movable', pie_chart_html)
        self.assertTrue(_all_theta_sum_is_360_degrees(pie_chart_html))
        self.assertTrue(_slices_are_well_placed(pie_chart_html))


class TestMarkArcError(unittest.TestCase):
    """Pie chart error tests."""
    def test_load_data_url_error(self):
        """Pie chart load data url error."""
        with self.assertRaises(IOError) as error:
            aframexr.Chart(NON_EXISTING_URL_DATA).mark_arc().encode(color='model', theta='sales').to_html()

        self.assertEqual(str(error.exception), f"Could not load data from URL: {NON_EXISTING_URL_DATA.url}.")

    def test_local_file_does_not_exist(self):
        """Pie chart error when the local file does not exist."""
        with self.assertRaises(IOError) as error:
            aframexr.Chart(NON_EXISTING_LOCAL_PATH).mark_arc().encode(color='model', theta='sales').to_html()

        self.assertRegex(str(error.exception), r'Local file .* was not found')

    def test_bad_file_format(self):
        """Pie chart error when the file format is incorrect."""
        with open(BAD_FILE_FORMAT.url, 'w'):
            pass  # Create a bad format temporal file

        try:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(BAD_FILE_FORMAT).mark_arc().encode(color='model', theta='sales').to_html()

            self.assertIn('Unsupported file type: ', str(error.exception))

        finally:
            from pathlib import Path
            Path(BAD_FILE_FORMAT.url).unlink()  # Remove the temporal file

    def test_file_is_empty(self):
        """Pie chart error when the file is empty."""
        with open(EMPTY_FILE.url, 'w'):
            pass  # Create an empty temporal file

        try:
            with self.assertRaises(IOError) as error:
                aframexr.Chart(EMPTY_FILE).mark_arc().encode(color='model', theta='sales').to_html()

            self.assertIn('Error when processing data. Error: ', str(error.exception))

        finally:
            from pathlib import Path
            Path(EMPTY_FILE.url).unlink()  # Remove the temporal file

    def test_color_encoding_not_nominal_error(self):
        """Pie chart error when color encoding is not nominal."""
        with self.assertRaises(ValueError) as error:
            aframexr.Chart(DATA).mark_arc().encode(color='doors', theta='sales').to_html()  # Column doors, quantitative
        self.assertEqual(
            str(error.exception),
            ERROR_MESSAGES['COLOR_ENCODING_NOT_NOMINAL'].format(color_encoding='quantitative')
        )

    def test_theta_encoding_not_quantitative_error(self):
        """Pie chart error when theta encoding is not quantitative."""
        with self.assertRaises(ValueError) as error:
            aframexr.Chart(DATA).mark_arc().encode(color='model', theta='color').to_html()  # Column color, nominal
        self.assertEqual(str(error.exception), 'Theta-channel data must be quantitative')

    def test_depth_error(self):
        """Pie chart error when depth is incorrect."""
        for d in NOT_GREATER_THAN_0_NUMBERS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, depth=d).mark_arc().encode(color='model', theta='sales').to_html()
            self.assertEqual(
                str(error.exception),
                ERROR_MESSAGES['POSITIVE_NUMBER'].format(param_name='depth')
            )

    def test_position_error(self):
        """Pie chart position error."""
        for p in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_arc().encode(color='model', theta='sales').to_html()
            self.assertEqual(
                str(error.exception),
                ERROR_MESSAGES['NOT_3_AXES_POSITION_OR_ROTATION'].format(pos_or_rot='position', pos_or_rot_value=p)
            )

        for p in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_arc().encode(color='model', theta='sales').to_html()
            self.assertEqual(str(error.exception), 'The position values must be numeric.')

    def test_rotation_error(self):
        """Pie chart rotation error."""
        for r in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_arc().encode(color='model', theta='sales').to_html()
            self.assertEqual(str(error.exception), ERROR_MESSAGES['NOT_3_AXES_POSITION_OR_ROTATION'].format(
                pos_or_rot='rotation', pos_or_rot_value=r
            ))

        for r in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_arc().encode(color='model', theta='sales').to_html()
            self.assertEqual(str(error.exception), 'The rotation values must be numeric.')

    def test_radius_error(self):
        """Pie chart radius error."""
        for r in NOT_GREATER_THAN_0_RADIUS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_arc(radius=r).encode(color='model', theta='sales')
            self.assertEqual(str(error.exception), ERROR_MESSAGES['POSITIVE_NUMBER'].format(param_name='radius'))

    def test_encoding_error(self):
        """Pie chart encoding error."""
        for e in NON_EXISTING_MARK_ARC_ENCODINGS:
            with self.assertRaises(KeyError) as error:
                pie_chart = aframexr.Chart(DATA).mark_arc().encode(**e)
                pie_chart.to_html()
            self.assertIn('Data has no field ', str(error.exception))

        for e in NOT_VALID_MARK_ARC_ENCODINGS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_arc().encode(**e).to_html()
            self.assertIn(
                str(error.exception),
                [
                    ERROR_MESSAGES['PARAM_NOT_SPECIFIED_IN_MARK_ARC'].format(param='color'),
                    ERROR_MESSAGES['PARAM_NOT_SPECIFIED_IN_MARK_ARC'].format(param='theta')
                ]
            )

    def test_encoding_error_invalid_encoding_type(self):
        """Pie chart encoding error with invalid encoding type."""
        with self.assertRaises(ValueError) as error:
            bad_encoding_type = 'BAD_ENCODING'
            aframexr.Chart(DATA).mark_arc().encode(color=f'model:{bad_encoding_type}', theta='sales').to_html()
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENCODING_TYPE'].format(encoding_type=bad_encoding_type))

    def test_encoding_error_not_encoded(self):
        """Pie chart encoding error. Encoding not in specifications."""
        with self.assertRaises(ValueError) as error:
            pie_chart = aframexr.Chart(DATA).mark_arc()
            pie_chart.to_html()
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENCODING_NOT_IN_SPECS'])

    def test_filter_warning(self):
        """Pie chart filter warning."""
        for f in WARNING_FILTER_EQUATIONS:
            with self.assertWarns(UserWarning) as warning:
                filt_chart = aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales').transform_filter(f)
                filt_chart.to_html()
            self.assertEqual(
                str(warning.warning),
                f'Data does not contain values for the filter: {FilterTransform.from_equation(f).to_dict()}'
            )

    def test_filter_error(self):
        """Pie chart filter error."""
        for f in SYNTAX_ERROR_FILTER_EQUATIONS:
            with self.assertRaises(SyntaxError) as error:
                filt_chart = aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales').transform_filter(f)
                filt_chart.to_html()
            self.assertEqual(str(error.exception), 'Incorrect syntax, must be datum.{field} {operator} {value}')

    def test_aggregate_error(self):
        """Pie chart aggregate error."""
        for a in NOT_VALID_AGGREGATES:
            with self.assertRaises(ValueError) as error:
                agg_chart = (aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales')
                             .transform_aggregate(new_field=f'{a}(sales)'))
                agg_chart.to_html()
            self.assertEqual(str(error.exception), ERROR_MESSAGES['AGGREGATE_OPERATION'].format(operation=a))

    def test_aggregate_error_not_defined_encoding_channels(self):
        """Pie chart aggregate error raised when encoding channels are not defined in aggregation."""
        with self.assertRaises(ValueError) as error:
            agg_chart = (aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales')
                         .transform_aggregate(mean_sales='mean(sales)', groupby=['model']))
            agg_chart.to_html()

        self.assertRegex(str(error.exception), r'Encoding channel\(s\) .* must be defined in aggregate .*'
                                               r', otherwise that fields will disappear.')

    def test_environment_error(self):
        """Scene creation with invalid personalized environment."""
        environment = 'bad_environment'
        pie_chart = aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales')
        with self.assertRaises(ValueError) as error:
            # noinspection PyTypeChecker
            pie_chart.to_html(environment=environment)
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENVIRONMENT'].format(environment=environment))



"""AframeXR filters"""

import polars as pl
from abc import ABC, abstractmethod
from polars import DataFrame

from ..utils.validators import AframeXRValidator


OPERATOR_MAP: dict[str, type['FilterTransform']] = {}  # Operator map, classes are added at the end of this file


class FilterTransform(ABC):
    """FilterTransform base class."""
    _operator: str = ''
    _magic_method: str = ''  # Must be defined by child classes with its method (e.g. __eq__)

    @abstractmethod
    def __init__(self, field: str, value: str | float):
        self.field = field
        self.value = value

    # Exporting equation formats
    @abstractmethod
    def to_dict(self):
        """Returns a dictionary about the equation of the filter with the syntaxis of the JSON specifications."""

    # Creating filters
    @staticmethod
    def from_equation(equation: str):
        AframeXRValidator.validate_type('equation', equation, str)

        parts = equation.split()
        if len(parts) != 3:
            raise SyntaxError('Incorrect syntax, must be datum.{field} {operator} {value}')

        field, op, value = parts

        try:
            value = float(value)
        except ValueError:
            pass

        if op in OPERATOR_MAP:
            if not field.startswith('datum.'):
                raise SyntaxError('Incorrect syntax, must be datum.{field} {operator} {value}')
            return OPERATOR_MAP[op](field.removeprefix('datum.'), value)
        else:
            raise ValueError(f'There is no filter for equation: {equation}')

    @staticmethod
    def from_dict(filter_specs: dict):
        """
        Creates a child filter object from the given filter's specifications.

        Parameters
        ----------
        filter_specs : dict
            Filter specifications.

        Raises
        ------
        TypeError
            If equation is not a dictionary.
        ValueError
            If the specifications of the filter is not correct.
        """
        AframeXRValidator.validate_type('equation', filter_specs, dict)

        if 'equal' in filter_specs:  # Equation is of type field == value
            return FieldEqualPredicate(filter_specs['field'], filter_specs['equal'])
        if 'gt' in filter_specs:  # Equation is of type field > value
            return FieldGTPredicate(filter_specs['field'], filter_specs['gt'])
        if 'lt' in filter_specs:  # Equation is of type field < value
            return FieldLTPredicate(filter_specs['field'], filter_specs['lt'])
        else:
            raise ValueError(f'There is no filter for specifications: {filter_specs}')

    # Filter data
    def get_filtered_data(self, data: DataFrame) -> DataFrame:
        """Filters and returns the data."""
        if not self._magic_method:  # pragma: no cover
            raise RuntimeError(f'Unreachable code. Magic method was not defined in {self.__class__.__name__} class')

        try:
            condition = getattr(pl.col(self.field), self._magic_method)(self.value)
            filtered_data = data.filter(condition)
        except pl.exceptions.ColumnNotFoundError:
            raise KeyError(f'Data has no field "{self.field}".')
        return filtered_data


class FieldEqualPredicate(FilterTransform):
    """Equal predicate filter class."""

    def __init__(self, field: str, equal: str | float):
        self._operator = '=='
        self._magic_method = '__eq__'  # Magic method
        super().__init__(field, equal)

    def to_dict(self):
        return {'field': self.field, 'equal': self.value}


class FieldGTPredicate(FilterTransform):
    """Greater than predicate filter class."""

    def __init__(self, field: str, gt: float):
        self._operator = '>'
        self._magic_method = '__gt__'  # Magic method
        super().__init__(field, gt)

    def to_dict(self):
        return {'field': self.field, 'gt': self.value}


class FieldLTPredicate(FilterTransform):
    """Lower than predicate filter class."""

    def __init__(self, field: str, lt: float):
        self._operator = '<'
        self._magic_method = '__lt__'  # Magic method
        super().__init__(field, lt)

    def to_dict(self):
        return {'field': self.field, 'lt': self.value}


# Add classes to OPERATOR_MAP
OPERATOR_MAP.update({
    '==': FieldEqualPredicate,
    '>': FieldGTPredicate,
    '<': FieldLTPredicate,
})



import aframexr
import unittest

from aframexr.utils.constants import AVAILABLE_ENVIRONMENTS, ERROR_MESSAGES
from .constants import *


class TestTorusOK(unittest.TestCase):
    """Tests for torus creation."""

    def test_simple(self):
        """Simple torus creation."""
        torus = aframexr.Torus()
        torus.to_html()

    def test_from_dict(self):
        """Torus creation using form_dict() method."""
        torus = aframexr.Torus.from_dict(aframexr.Torus().to_dict())
        torus.to_html()

    def test_from_json(self):
        """Torus creation using form_json() method."""
        torus = aframexr.Torus.from_json(aframexr.Torus().to_json())
        torus.to_html()

    def test_color(self):
        """Torus with specific color creation."""
        for c in SIMPLE_ELEMENTS_COLORS:
            torus = aframexr.Torus(color=c)
            torus.to_html()

    def test_position(self):
        """Torus with specific position creation."""
        for p in POSITIONS + POSITION_FORMATS:
            torus = aframexr.Torus(position=p)
            torus.to_html()

    def test_radius(self):
        """Torus with specific radius creation."""
        for r in RADIUS:
            torus = aframexr.Torus(radius=r)
            torus.to_html()

    def test_radius_tubular(self):
        """Torus with specific radius tubular creation."""
        for r in RADIUS:
            torus = aframexr.Torus(radius_tubular=r)
            torus.to_html()

    def test_rotation(self):
        """Torus with specific rotation creation."""
        for r in ROTATIONS + ROTATION_FORMATS:
            torus = aframexr.Torus(rotation=r)
            torus.to_html()

    def test_all_parameters(self):
        """Torus creation with all parameters specified."""
        for c, p, rad, rt, rot in zip(SIMPLE_ELEMENTS_COLORS, POSITIONS, RADIUS, RADIUS, ROTATIONS):
            torus = aframexr.Torus(color=c, position=p, radius=rad, radius_tubular=rt, rotation=rot)
            torus.to_html()

    def test_concatenation(self):
        """Concatenation of simple tetrahedron."""
        concatenated_chart = aframexr.Torus(position=CONCATENATION_POSITIONS[0])
        for p in CONCATENATION_POSITIONS[1:]:
            concatenated_chart += aframexr.Torus(position=p)
        concatenated_chart.to_html()

    def test_environment(self):
        """Scene creation with personalized environment."""
        for e in AVAILABLE_ENVIRONMENTS:
            aframexr.Torus().to_html(environment=e)

    def test_movable(self):
        """Movable torus creation."""
        html = aframexr.Torus().movable().to_html()
        self.assertIn('movable', html)


class TestTorusERROR(unittest.TestCase):
    """Tests for torus creation error."""

    def test_color_error(self):
        """Torus with specific color creation error."""
        not_an_string = ['not a color']
        with self.assertRaises(TypeError) as error:
            # noinspection PyTypeChecker
            torus = aframexr.Torus(color=not_an_string)
            torus.to_html()
        self.assertEqual(
            str(error.exception),
            ERROR_MESSAGES['TYPE'].format(
                param_name='specs.color', expected_type=str.__name__, current_type=type(not_an_string).__name__
            )
        )

    def test_radius_error(self):
        """Torus with specific radius creation error."""
        for r in NOT_GREATER_THAN_0_RADIUS:
            with self.assertRaises(ValueError) as error:
                torus = aframexr.Torus(radius=r)
                torus.to_html()
            self.assertEqual(str(error.exception), ERROR_MESSAGES['POSITIVE_NUMBER'].format(param_name='specs.radius'))

    def test_radius_tubular_error(self):
        """Torus with specific radius tubular creation error."""
        for r in NOT_GREATER_THAN_0_RADIUS:
            with self.assertRaises(ValueError) as error:
                torus = aframexr.Torus(radius_tubular=r)
                torus.to_html()
            self.assertEqual(
                str(error.exception),
                ERROR_MESSAGES['POSITIVE_NUMBER'].format(param_name='specs.radius_tubular')
            )

    def test_environment_error(self):
        """Scene creation with invalid personalized environment."""
        environment = 'bad_environment'
        torus = aframexr.Torus()
        with self.assertRaises(ValueError) as error:
            # noinspection PyTypeChecker
            torus.to_html(environment=environment)
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENVIRONMENT'].format(environment=environment))



"""AframeXR validators"""

from typing import Literal

from .constants import (
    AVAILABLE_AGGREGATES, AVAILABLE_ENCODING_TYPES, AVAILABLE_ENVIRONMENTS, AVAILABLE_MARKS, ERROR_MESSAGES
)
from .element_creator import CREATOR_MAP


def _validate_3_axes_numerical_values(param_name: str, param_value: str) -> None:
    """Raises TypeError or ValueError if parameter is not a string of 3 numerical values."""
    AframeXRValidator.validate_type(param_name, param_value, str)

    pos_axes = param_value.strip().split()
    if len(pos_axes) != 3:
        raise ValueError(ERROR_MESSAGES['NOT_3_AXES_POSITION_OR_ROTATION'].format(
            pos_or_rot=param_name, pos_or_rot_value=param_value)
        )
    for axis in pos_axes:
        try:
            float(axis)
        except ValueError:
            raise ValueError(f'The {param_name} values must be numeric.')


def _validate_align(align: Literal['center', 'left', 'right']) -> None:
    """Raises TypeError or ValueError if align is invalid."""
    AframeXRValidator.validate_type('specs.align', align, str)
    if align not in ('center', 'left', 'right'):
        raise ValueError(ERROR_MESSAGES['ALIGN'].format(align=align))


def _validate_data(data: dict) -> None:
    """Raises TypeError or ValueError if data is invalid."""
    AframeXRValidator.validate_type('specs.data', data, dict)
    if 'values' in data and 'url' in data:
        raise ValueError(ERROR_MESSAGES['DATA_WITH_VALUES_AND_URL_IN_SPECS'])

    if 'values' in data:
        values = data['values']
        AframeXRValidator.validate_type('specs.data.values', values, list)
        if not all(isinstance(v, dict) for v in values):
            raise TypeError(ERROR_MESSAGES['NOT_ALL_DATA_VALUES_ARE_DICT'])

    elif 'url' in data:
        AframeXRValidator.validate_type('specs.data.url', data['url'], str)

    else:
        raise ValueError(ERROR_MESSAGES['DATA_WITH_NOT_VALUES_NEITHER_URL_IN_SPECS'])


def _validate_element(element: str) -> None:
    """Raises TypeError or ValueError if element is invalid."""
    AframeXRValidator.validate_type('specs.element', element, str)

    if element not in CREATOR_MAP.keys():
        raise ValueError(ERROR_MESSAGES['ELEMENT_TYPE'].format(element=element))


def _validate_encoding(encoding: dict):
    """Raises TypeError or ValueError if encoding is invalid."""
    AframeXRValidator.validate_type('specs.encoding', encoding, dict)

    if not all(isinstance(e, dict) for e in encoding.values()):
        raise TypeError(ERROR_MESSAGES['NOT_ALL_ENCODINGS_ARE_DICT'])


def _validate_mark(mark: str | dict) -> None:
    """Raises TypeError or ValueError if "mark" is invalid."""
    AframeXRValidator.validate_type('specs.mark', mark, (str, dict))

    mark_type = mark.get('type') if isinstance(mark, dict) else mark
    if mark_type not in AVAILABLE_MARKS:
        raise ValueError(ERROR_MESSAGES['MARK_TYPE'].format(mark_type=mark_type))


def _validate_mark_encoding(mark: str | dict, encoding: dict) -> None:
    """
    Raises ValueError if "mark" and "encoding" combination is invalid.

    Notes
    -----
    Assuming that mark and encoding are valid, as this method is called after mark and encoding validation.
    """
    mark_type = mark.get('type') if isinstance(mark, dict) else mark

    if mark_type in {'bar', 'point'} \
            and sum([encoding.get('x') is not None, encoding.get('y') is not None, encoding.get('z') is not None]) < 2:
        raise ValueError(ERROR_MESSAGES['LESS_THAN_2_XYZ_ENCODING'])
    if mark_type == 'arc' and (encoding.get('color') is None or encoding.get('theta') is None):
        if encoding.get('color') is None:
            raise ValueError(ERROR_MESSAGES['PARAM_NOT_SPECIFIED_IN_MARK_ARC'].format(param='color'))
        if encoding.get('theta') is None:
            raise ValueError(ERROR_MESSAGES['PARAM_NOT_SPECIFIED_IN_MARK_ARC'].format(param='theta'))


def _validate_params(params: list) -> None:
    """Raises TypeError or ValueError if params are invalid."""
    AframeXRValidator.validate_type('specs.params', params, list)

    for p in params:
        AframeXRValidator.validate_type('specs.params.param', p, dict)
        if not 'name' in p:
            raise ValueError(ERROR_MESSAGES['NAME_NOT_IN_PARAM'])


def _validate_transform(transform: list[dict]) -> None:
    """Raises TypeError or ValueError if transform is invalid."""
    AframeXRValidator.validate_type('specs.transform', transform, list)

    for t in transform:
        AframeXRValidator.validate_type(f'specs.transform.{t}', t, dict)

        if t.get('filter'):
            AframeXRValidator.validate_type('specs.transform.filter', t['filter'], dict)
            if 'param' in t['filter']:
                AframeXRValidator.validate_type('specs.transform.filter.param', t['filter']['param'], str)

        elif t.get('aggregate'):
            AframeXRValidator.validate_type('specs.transform.aggregate', t['aggregate'], list)
            for agg in t['aggregate']:
                if 'op' not in agg:
                    raise ValueError(ERROR_MESSAGES['AGGREGATE_OPERATION_NOT_IN_AGGREGATE'])
                AframeXRValidator.validate_aggregate_operation(agg['op'])

        else:
            raise ValueError(ERROR_MESSAGES['TRANSFORM_TYPE'].format(transform_type=t))


class AframeXRValidator:
    """AframeXR validator class."""

    @staticmethod
    def validate_aggregate_operation(aggregate_operation: str) -> None:
        """Raises TypeError or ValueError if aggregate operation is invalid."""
        AframeXRValidator.validate_type('aggregate operation', aggregate_operation, str)
        if aggregate_operation not in AVAILABLE_AGGREGATES:
            raise ValueError(ERROR_MESSAGES['AGGREGATE_OPERATION'].format(operation=aggregate_operation))

    @staticmethod
    def validate_chart_specs(specs: dict) -> None:
        """Raises ValueError if chart specifications are invalid."""
        AframeXRValidator.validate_type('specifications', specs, dict)
        if 'concat' in specs:
            charts = specs['concat']
            AframeXRValidator.validate_type('specs.concat', charts, list)
            for chart_specs in charts:  # There are several charts in the specifications
                AframeXRValidator.validate_chart_specs(chart_specs)  # Validate each chart specification
            return

        if 'mark' in specs and 'element' in specs:
            raise ValueError(ERROR_MESSAGES['MARK_AND_ELEMENT_IN_SPECS'])

        elif 'mark' in specs:  # Chart
            if 'data' not in specs:
                raise ValueError(ERROR_MESSAGES['DATA_NOT_IN_SPECS'])
            _validate_data(specs['data'])
            _validate_mark(specs['mark'])

            if 'encoding' not in specs:
                raise ValueError(ERROR_MESSAGES['ENCODING_NOT_IN_SPECS'])
            _validate_encoding(specs['encoding'])

            _validate_mark_encoding(specs['mark'], specs['encoding'])

            if 'params' in specs:
                _validate_params(specs['params'])

        elif 'element' in specs:  # Single element
            _validate_element(specs['element'])

            if 'align' in specs:
                _validate_align(specs['align'])
            if 'color' in specs:
                AframeXRValidator.validate_type('specs.color', specs['color'], str)
            if 'radius' in specs:
                AframeXRValidator.validate_positive_number('specs.radius', specs['radius'])
            if 'radius_bottom' in specs:
                AframeXRValidator.validate_positive_number('specs.radius_bottom', specs['radius_bottom'])
            if 'radius_top' in specs:
                AframeXRValidator.validate_positive_number('specs.radius_top', specs['radius_top'])
            if 'radius_tubular' in specs:
                AframeXRValidator.validate_positive_number('specs.radius_tubular', specs['radius_tubular'])
            if 'scale' in specs:
                _validate_3_axes_numerical_values('specs.scale', specs['scale'])
            if 'src' in specs:
                AframeXRValidator.validate_type('specs.src', specs['src'], str)
            if 'value' in specs:
                AframeXRValidator.validate_type('specs.value', specs['value'], str)

        else:
            raise ValueError(ERROR_MESSAGES['MARK_AND_ELEMENT_NOT_IN_SPECS'])

        if 'position' in specs:
            AframeXRValidator.validate_type('specs.position', specs['position'], str)
            _validate_3_axes_numerical_values('position', specs['position'])

        if 'rotation' in specs:
            AframeXRValidator.validate_type('specs.rotation', specs['rotation'], str)
            _validate_3_axes_numerical_values('rotation', specs['rotation'])

        if 'environment' in specs:
            AframeXRValidator.validate_type('specs.environment', specs['environment'], str)
            if specs['environment'] not in AVAILABLE_ENVIRONMENTS:
                raise ValueError(ERROR_MESSAGES['ENVIRONMENT'].format(environment=specs['environment']))

        if 'depth' in specs:
            AframeXRValidator.validate_positive_number('depth', specs['depth'])

        if 'height' in specs:
            AframeXRValidator.validate_positive_number('height', specs['height'])

        if 'width' in specs:
            AframeXRValidator.validate_positive_number('width', specs['width'])

        if 'transform' in specs:
            _validate_transform(specs['transform'])

    @staticmethod
    def validate_encoding_type(encoding_type: str) -> None:
        """Raises TypeError if encoding type is invalid."""
        if encoding_type not in AVAILABLE_ENCODING_TYPES:
            raise ValueError(ERROR_MESSAGES['ENCODING_TYPE'].format(encoding_type=encoding_type))

    @staticmethod
    def validate_positive_number(name: str, value: float | int):
        """Raises TypeError or ValueError if value is not greater than 0."""
        AframeXRValidator.validate_type(name, value, (float, int))
        if value <= 0:
            raise ValueError(ERROR_MESSAGES['POSITIVE_NUMBER'].format(param_name=name))

    @staticmethod
    def validate_type(param_name: str, param, types: type | tuple[type, ...]) -> None:
        """Raises TypeError if type(param) is not in types."""
        if not isinstance(param, types):
            if isinstance(types, tuple):
                expected = ' or '.join(t.__name__ for t in types)
            else:
                expected = types.__name__
            raise TypeError(ERROR_MESSAGES['TYPE'].format(
                param_name=param_name, expected_type=expected, current_type=type(param).__name__)
            )



import aframexr
import unittest

from aframexr.api.filters import FilterTransform
from aframexr.utils.constants import AVAILABLE_ENVIRONMENTS, ERROR_MESSAGES
from tests.constants import *  # Constants used for testing


class TestMarkLineOK(unittest.TestCase):
    """Line chart OK tests."""
    def test_simple(self):
        """Simple line chart creation."""
        line_chart = aframexr.Chart(DATA).mark_line().encode(x='model', y='sales')
        line_chart.to_html()

    def test_simple_with_pandas_not_installed(self):
        """Simple line chart creation without having pandas installed."""
        import sys
        import importlib
        from unittest import mock

        sys.modules.pop('aframexr', None)
        sys.modules.pop('aframexr.api.components', None)

        with mock.patch.dict(sys.modules, {'pandas': None}):
            import aframexr
            importlib.reload(aframexr)

            line_chart = aframexr.Chart(DATA).mark_line().encode(x='model', y='sales')
            line_chart.to_html()

            import aframexr.api.components as components
            self.assertIsNone(components.pd)
            self.assertIs(components.DataFrame, object)

    def test_from_json(self):
        """Line chart using from_json() method creation."""
        json_string = aframexr.Chart(DATA).mark_line().encode(x='model', y='sales').to_json()
        line_chart = aframexr.Chart.from_json(json_string)
        line_chart.to_html()

        self.assertEqual(line_chart.to_json(), json_string)

    def test_data_format(self):
        """Line chart changing data format creation."""
        for d in DATA_FORMATS:
            line_chart = aframexr.Chart(d).mark_line().encode(x='model', y='sales')
            line_chart.to_html()

    def test_position(self):
        """Line chart changing position creation."""
        for p in POSITIONS:
            line_chart = aframexr.Chart(DATA, position=p).mark_line().encode(x='model', y='sales')
            line_chart.to_html()

    def test_position_format(self):
        """Line chart changing position format creation."""
        for p in POSITION_FORMATS:
            line_chart = aframexr.Chart(DATA, position=p).mark_line().encode(x='model', y='sales')
            line_chart.to_html()

    def test_rotation(self):
        """Line chart changing rotation creation."""
        for r in ROTATIONS:
            line_chart = aframexr.Chart(DATA, rotation=r).mark_line().encode(x='model', y='sales')
            line_chart.to_html()

    def test_rotation_format(self):
        """Line chart changing rotation format creation."""
        for r in ROTATION_FORMATS:
            line_chart = aframexr.Chart(DATA, rotation=r).mark_line().encode(x='model', y='sales')
            line_chart.to_html()

    def test_height(self):
        """Line chart changing height creation."""
        for h in MARK_BAR_POINT_HEIGHTS_WIDTHS:
            line_chart = aframexr.Chart(DATA, height=h).mark_line().encode(x='model', y='sales')
            line_chart.to_html()

    def test_width(self):
        """Line chart changing width creation."""
        for w in MARK_BAR_POINT_HEIGHTS_WIDTHS:
            line_chart = aframexr.Chart(DATA, width=w).mark_line().encode(x='model', y='sales')
            line_chart.to_html()

    def test_title(self):
        """Line chart with title creation."""
        line_chart = aframexr.Chart(DATA, title='Title').mark_line().encode(x='model', y='sales')
        line_chart.to_html()

    def test_color(self):
        """Line chart changing color creation."""
        for c in SIMPLE_ELEMENTS_COLORS:
            line_chart = aframexr.Chart(DATA).mark_line(color=c).encode(x='model', y='sales')
            line_chart.to_html()

    def test_point_markers(self):
        """Line chart using point markers creation."""
        line_chart = aframexr.Chart(DATA).mark_line(point=True).encode(x='model', y='sales')
        line_chart.to_html()

    def test_encoding(self):
        """Line chart changing encoding creation."""
        for e in MARK_BAR_ENCODINGS:
            line_chart = aframexr.Chart(DATA).mark_line().encode(**e)
            line_chart.to_html()

    def test_encoding_with_no_Y_axis_displayed(self):
        """Line chart creation with no Y-axis displayed."""
        line_chart = aframexr.Chart(DATA).mark_line().encode(x='model', y=aframexr.Y('sales', axis=None))
        line_chart.to_html()

    def test_filter(self):
        """Line chart changing filter creation."""
        for eq in FILTER_EQUATIONS:
            for f in [eq, FilterTransform.from_equation(eq)]:  # Filter using equation and using FilterTransform object
                line_chart = aframexr.Chart(DATA).mark_line().encode(x='model', y='sales').transform_filter(f)
                line_chart.to_html()

    def test_several_filters(self):
        """Line chart with several filters' creation."""
        line_chart = (aframexr.Chart(DATA).mark_line().encode(x='model', y='sales')
                      .transform_filter(SEVERAL_FILTER_EQUATIONS[0]))
        for eq in SEVERAL_FILTER_EQUATIONS[1:]:
            line_chart.transform_filter(eq)

        line_chart.to_html()

    def test_aggregate(self):
        """Line chart changing aggregates creation."""
        for a in AGGREGATES:
            line_chart = (aframexr.Chart(DATA).mark_line().encode(x='model', y='new_field')
                          .transform_aggregate(new_field=f'{a}(sales)'))
            line_chart.to_html()

            line_chart_2 = aframexr.Chart(DATA).mark_line().encode(x='model', y=f'{a}(sales)')
            line_chart_2.to_html()

    def test_aggregate_position_rotation_size_height_width_filter(self):
        """Line chart changing position, rotation size, height, width and filter creation."""
        for a, p, r, po, h, w, f in zip(AGGREGATES, POSITIONS, ROTATIONS, (True, False),
                                       MARK_BAR_POINT_HEIGHTS_WIDTHS, MARK_BAR_POINT_HEIGHTS_WIDTHS,
                                       FILTER_EQUATIONS):
            line_chart = (aframexr.Chart(DATA, position=p, rotation=r, height=h, width=w).mark_line(point=po)
                          .encode(x='model', y='agg').transform_filter(f).transform_aggregate(agg=f'{a}(sales)'))
            line_chart.to_html()

    def test_concatenation(self):
        """Line chart concatenation creation."""
        concatenated_chart = (aframexr.Chart(DATA, position=CONCATENATION_POSITIONS[0]).mark_line()
                              .encode(x='model', y='sales'))
        for pos in CONCATENATION_POSITIONS[1:]:
            concatenated_chart += aframexr.Chart(DATA, position=pos).mark_line().encode(x='model', y='sales')

        concatenated_chart.to_html()

    def test_environment(self):
        """Scene creation with personalized environment."""
        for e in AVAILABLE_ENVIRONMENTS:
            line_chart = aframexr.Chart(DATA).mark_line().encode(x='model', y='sales')
            line_chart.to_html(environment=e)

    def test_save(self):
        """Line chart saving."""
        import tempfile
        from pathlib import Path

        with tempfile.TemporaryDirectory() as tmpdir:
            temp_html_file_path = Path(tmpdir) / "test.html"
            temp_json_file_path = Path(tmpdir) / "test.json"

            line_chart = aframexr.Chart(DATA).mark_line().encode(x='model', y='sales')
            line_chart.save(str(temp_html_file_path))
            line_chart.save(str(temp_json_file_path))

            self.assertTrue(temp_html_file_path.exists())
            self.assertTrue(temp_json_file_path.exists())

    def test_properties(self):
        """Line chart properties definition."""
        line_chart = aframexr.Chart().mark_line().encode(x='model', y='sales')
        for p, r in zip(POSITIONS, ROTATIONS):
            line_chart_2 = line_chart.properties(data=DATA, position=p, rotation=r)
            line_chart_2.to_html()

    def test_movable(self):
        """Movable line chart creation."""
        line_chart_html = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales').movable().to_html()
        self.assertIn('movable', line_chart_html)


class TestMarkLineError(unittest.TestCase):
    """Line chart error tests."""

    def test_load_data_url_error(self):
        """Line chart load data url error."""
        with self.assertRaises(IOError) as error:
            aframexr.Chart(NON_EXISTING_URL_DATA).mark_line().encode(x='model', y='sales').to_html()

        self.assertEqual(str(error.exception), f"Could not load data from URL: {NON_EXISTING_URL_DATA.url}.")

    def test_local_file_does_not_exist(self):
        """Line chart error when a local file does not exist."""
        with self.assertRaises(IOError) as error:
            aframexr.Chart(NON_EXISTING_LOCAL_PATH).mark_line().encode(x='model', y='sales').to_html()

        self.assertRegex(str(error.exception), r'Local file .* was not found')

    def test_bad_file_format(self):
        """Line chart error when the file format is incorrect."""
        with open(BAD_FILE_FORMAT.url, 'w'):
            pass  # Create a bad format temporal file

        try:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(BAD_FILE_FORMAT).mark_line().encode(x='model', y='sales').to_html()

            self.assertIn(f'Unsupported file type: ', str(error.exception))

        finally:
            from pathlib import Path
            Path(BAD_FILE_FORMAT.url).unlink()  # Remove the temporal file

    def test_file_is_empty(self):
        """Line chart error when the file is empty."""
        with open(EMPTY_FILE.url, 'w'):
            pass  # Create an empty temporal file

        try:
            with self.assertRaises(IOError) as error:
                aframexr.Chart(EMPTY_FILE).mark_line().encode(x='model', y='sales').to_html()

            self.assertIn('Error when processing data. Error: ', str(error.exception))

        finally:
            from pathlib import Path
            Path(EMPTY_FILE.url).unlink()  # Remove the temporal file

    def test_color_encoding_not_nominal_error(self):
        """Line chart error when color encoding is not nominal."""
        with self.assertRaises(ValueError) as error:
            aframexr.Chart(DATA).mark_line().encode(x='model', y='sales', color='doors').to_html()  # Doors quantitative
        self.assertEqual(
            str(error.exception),
            ERROR_MESSAGES['COLOR_ENCODING_NOT_NOMINAL'].format(color_encoding='quantitative')
        )

    def test_position_error(self):
        """Line chart position error."""
        for p in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_line().encode(x='model', y='sales').to_html()
            self.assertEqual(str(error.exception), ERROR_MESSAGES['NOT_3_AXES_POSITION_OR_ROTATION'].format(
                pos_or_rot='position', pos_or_rot_value=p
            ))

        for p in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_line().encode(x='model', y='sales').to_html()
            self.assertEqual(str(error.exception), 'The position values must be numeric.')

    def test_rotation_error(self):
        """Line chart rotation error."""
        for r in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_line().encode(x='model', y='sales').to_html()
            self.assertEqual(str(error.exception), ERROR_MESSAGES['NOT_3_AXES_POSITION_OR_ROTATION'].format(
                pos_or_rot='rotation', pos_or_rot_value=r
            ))

        for r in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_line().encode(x='model', y='sales').to_html()
            self.assertEqual(str(error.exception), 'The rotation values must be numeric.')

    def test_height_error(self):
        """Line chart height error."""
        for h in NOT_GREATER_THAN_0_NUMBERS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, height=h).mark_line().encode(x='model', y='sales').to_html()
            self.assertEqual(str(error.exception), ERROR_MESSAGES['POSITIVE_NUMBER'].format(param_name='height'))

    def test_width_error(self):
        """Line chart width error."""
        for w in NOT_GREATER_THAN_0_NUMBERS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, width=w).mark_line().encode(x='model', y='sales').to_html()
            self.assertEqual(str(error.exception), ERROR_MESSAGES['POSITIVE_NUMBER'].format(param_name='width'))

    def test_encoding_error(self):
        """Line chart encoding error."""
        for e in NON_EXISTING_MARK_BAR_POINT_ENCODINGS:
            with self.assertRaises(KeyError) as error:
                aframexr.Chart(DATA).mark_line().encode(**e).to_html()
            self.assertIn('Data has no field ', str(error.exception))

    def test_encoding_error_invalid_encoding_type(self):
        """Line chart encoding error with invalid encoding type."""
        with self.assertRaises(ValueError) as error:
            bad_encoding_type = 'BAD_ENCODING'
            aframexr.Chart(DATA).mark_line().encode(x=f'model:{bad_encoding_type}', y='sales').to_html()
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENCODING_TYPE'].format(encoding_type=bad_encoding_type))

    def test_encoding_error_not_encoded(self):
        """Line chart encoding error. Encoding not in specifications."""
        with self.assertRaises(ValueError) as error:
            aframexr.Chart(DATA).mark_line().to_html()
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENCODING_NOT_IN_SPECS'])

    def test_filter_warning(self):
        """Line chart filter warning."""
        for f in WARNING_FILTER_EQUATIONS:
            with self.assertWarns(UserWarning) as warning:
                aframexr.Chart(DATA).mark_line().encode(x='model', y='sales').transform_filter(f).to_html()
            self.assertEqual(
                str(warning.warning),
                f'Data does not contain values for the filter: {FilterTransform.from_equation(f).to_dict()}'
            )

    def test_filter_error(self):
        """Line chart filter error."""
        for f in SYNTAX_ERROR_FILTER_EQUATIONS:
            with self.assertRaises(SyntaxError) as error:
                aframexr.Chart(DATA).mark_line().encode(x='model', y='sales').transform_filter(f).to_html()
            self.assertEqual(str(error.exception), 'Incorrect syntax, must be datum.{field} {operator} {value}')

    def test_aggregate_error(self):
        """Line chart aggregate error."""
        for a in NOT_VALID_AGGREGATES:
            with self.assertRaises(ValueError) as error:
                (aframexr.Chart(DATA).mark_line().encode(x='model', y='sales')
                 .transform_aggregate(new_field=f'{a}(sales)').to_html())
            self.assertEqual(str(error.exception), ERROR_MESSAGES['AGGREGATE_OPERATION'].format(operation=a))

    def test_aggregate_error_not_defined_encoding_channels(self):
        """Line chart aggregate error raised when encoding channels are not defined in aggregation."""
        with self.assertRaises(ValueError) as error:
            (aframexr.Chart(DATA).mark_line().encode(x='model', y='sales')
             .transform_aggregate(mean_sales='mean(sales)', groupby=['model'])).to_html()

        self.assertRegex(str(error.exception), r'Encoding channel\(s\) .* must be defined in aggregate .*'
                                               r', otherwise that fields will disappear.')

    def test_environment_error(self):
        """Scene creation with invalid personalized environment."""
        environment = 'bad_environment'
        line_chart = aframexr.Chart(DATA).mark_line().encode(x='model', y='sales')
        with self.assertRaises(ValueError) as error:
            # noinspection PyTypeChecker
            line_chart.to_html(environment=environment)
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENVIRONMENT'].format(environment=environment))



"""AframeXR charts creator"""

import polars as pl
import warnings

from itertools import cycle, islice
from polars import DataFrame, Series
from polars.datatypes.group import NUMERIC_DTYPES
from typing import Literal

from .axis_creator import AxisCreator
from .constants import *
from .element_creator import (
    BoxCreator, CylinderCreator, ElementCreator, PlaneCreator, SphereCreator, TextCreator, LineCreator
)

CREATOR_MAP: dict[str, type['ChartCreator']] = {}  # Creator map of charts, classes are added at the end of this file


def _calculate_point_radius(point_volume: float) -> float:
    return (3 * point_volume / (4 * 3.1416)) ** (1 / 3)


def _translate_dtype_into_encoding(dtype: pl.DataType) -> str:
    """Translates and returns the encoding for a given data type."""

    if dtype in NUMERIC_DTYPES:
        encoding_type = 'quantitative'
    elif dtype in (pl.String, pl.Categorical):
        encoding_type = 'nominal'
    else:
        raise ValueError(f'Unknown dtype: {dtype}.')
    return encoding_type


class ChartCreator:
    """Chart creator base class"""

    def __init__(self, chart_specs: dict):
        base_position = chart_specs.get('position', DEFAULT_CHART_POS)
        [self._base_x, self._base_y, self._base_z] = [float(pos) for pos in base_position.split()]  # Base position
        self._encoding = chart_specs.get('encoding')  # Encoding and parameters of the chart
        rotation = chart_specs.get('rotation', DEFAULT_CHART_ROTATION)  # Rotation of the chart
        [self._x_rotation, self._y_rotation, self._z_rotation] = [float(rot) for rot in rotation.split()]
        self._params = chart_specs.get('params', [])  # Metadata parameters
        self._process_params()
        self._raw_data = DataFrame(chart_specs['data']['values'])

        self._elements_colors_all = chart_specs['mark'].get('color', DEFAULT_ELEMENTS_COLOR_IN_CHART) \
            if isinstance(chart_specs['mark'], dict) else DEFAULT_ELEMENTS_COLOR_IN_CHART
        self._color_data: Series | None = None
        self._color_encoding: str = ''

        self._chart_depth = chart_specs.get('depth')  # Maximum depth of the chart

        self._title = chart_specs.get('title')
        # Each self._{channel} attributes must be named by child classes

    def _add_selection_to_specs(self, specs: dict) -> None:
        if not getattr(self, "_selection", None):
            return

        param_name = self._selection['name']
        select_fields = self._selection['fields']

        expressions = [pl.lit(param_name)] + [
            pl.col(f).cast(pl.Utf8).fill_null("")
            for f in select_fields
        ]
        specs['activates_param'] = self._raw_data.select(
            pl.concat_str(expressions, separator="__")
        ).to_series().to_list()

    def _process_channels(self, *channels_name: str):
        """
        Process and stores the necessary channels' information.
        Must have defined self._{ch}_data and self._{ch}_encoding.
        """
        if self._raw_data.is_empty():
            return

        for ch in channels_name:
            if self._encoding.get(ch):
                channel_encoding = self._encoding[ch]
                field = channel_encoding['field']  # Field of the channel
                try:
                    data = self._raw_data[field]

                    detected_encoding = _translate_dtype_into_encoding(data.dtype)
                    user_encoding = channel_encoding.get('type', detected_encoding)
                    setattr(self, f'_{ch}_encoding', user_encoding)

                    # Set value for self._{ch}_data
                    if user_encoding == 'nominal' and detected_encoding == 'quantitative':
                        setattr(self, f'_{ch}_data', data.cast(pl.String))
                    else:
                        setattr(self, f'_{ch}_data', data)
                except pl.exceptions.ColumnNotFoundError:
                    raise KeyError(f'Data has no field "{field}" for {ch}-channel.')

    def _process_params(self):
        for p in self._params:
            if 'select' in p and p['select']['type'] == 'point':
                self._selection = {
                    'name': p['name'],
                    'fields': p['select'].get('fields', [])
                }

    def _set_elements_colors(self) -> Series:
        """Returns a Series of the color for each element composing the chart."""
        if self._color_encoding and self._color_encoding != 'nominal':
            raise ValueError(ERROR_MESSAGES['COLOR_ENCODING_NOT_NOMINAL'].format(color_encoding=self._color_encoding))

        if self._color_data is None:
            points_colors = pl.repeat(
                value=self._elements_colors_all,
                n=self._raw_data.height,  # Number of rows in data
                eager=True  # Returns a Series
            )
        else:
            unique_categories = self._color_data.unique(maintain_order=True).to_list()
            color_map = dict(zip(
                unique_categories,
                islice(cycle(AVAILABLE_COLORS), len(unique_categories))
            ))

            points_colors = self._color_data.replace(color_map)
        return points_colors.alias('color')

    @staticmethod
    def create_object(chart_type: str, chart_specs: dict):
        """Returns a ChartCreator instance of the specific chart type."""
        try:
            creator = CREATOR_MAP[chart_type]
        except KeyError:  # pragma: no cover (creator classes should be added at the end of this file)
            raise RuntimeError(f'Class for {chart_type} was not added to CREATOR_MAP')
        return creator(chart_specs)

    def get_relative_bottom_left_corner_position(self) -> str:
        """Returns the relative position for the bottom left corner of the chart."""
        # Only XYZAxisChannelChartCreator objects has that attribute, if not return '0 0 0'
        return getattr(self, '_relative_bottom_left_corner_position', '0 0 0')

    def get_axes_specs(self):  # pragma: no cover (get_axes_specs() must be implemented by child classes)
        raise RuntimeError('Unreachable code. Method get_axes_specs() must be implemented by child classes')

    def get_elements(self, filtered_by_params: bool) -> list[ElementCreator]:  # pragma: no cover
        raise RuntimeError('Unreachable code. Method get_elements() must be implemented by child classes')

    def get_group_specs(self) -> dict:
        """Returns a dictionary with the base specifications for the group of elements."""
        group_specs = {
            'position': f'{self._base_x} {self._base_y} {self._base_z}',
            'rotation': f'{self._x_rotation} {self._y_rotation} {self._z_rotation}'
        }

        return group_specs

    def get_legend_elements(self, filtered_by_params: bool = False) -> list[ElementCreator]:
        return []  # This method is redefined in charts that could have legend

    def get_title_elements(self, filtered_by_params: bool = False) -> list[ElementCreator]:
        if not self._title:
            return []

        # Relative title position
        if getattr(self, '_chart_width', None) is not None and getattr(self, '_chart_height', None) is not None:
            chart_height = getattr(self, '_chart_height')
            chart_width = getattr(self, '_chart_width')
            rotation = '0 0 0'
            title_position = f'{chart_width / 2} {chart_height + PLANE_TITLE_SEPARATION + PLANE_TITLE_HEIGHT / 2} 0'
        elif getattr(self, '_radius', None) is not None:  # Pie chart
            chart_height = chart_width = getattr(self, '_radius') * 2
            rotation = '90 0 0'  # Invert pie chart's rotation (-90 0 0)
            title_position = f'0 0 {chart_height / 2 + PLANE_TITLE_SEPARATION + PLANE_TITLE_HEIGHT / 2}'  # Z-axis
        else:  # pragma: no cover
            raise RuntimeError('Unreachable code')

        plane = PlaneCreator(
            {
                'position': title_position, 'rotation': rotation, 'height': PLANE_TITLE_HEIGHT, 'width': chart_width
            },
            filtered_by_params=filtered_by_params
        )
        text = TextCreator(
            {
                'position': title_position, 'rotation': rotation, 'value': self._title, 'color': TITLE_TEXT_COLOR,
                'align': 'center', 'scale': TITLE_TEXT_SCALE
            },
            filtered_by_params=filtered_by_params
        )

        return [plane, text]


# First-level subclasses of ChartCreator.
class XYZAxisChannelChartCreator(ChartCreator):
    """
    Chart creator base class for charts that have channels and XYZ axis.

    Notes
    -----
    XYZ-axes are processed instantly when creating this class or derivatives.
    """

    _AXIS_SIZE_MAP = {'x': 'width', 'y': 'height', 'z': 'depth'}

    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._chart_height = chart_specs.get('height')  # Maximum height of the chart
        self._chart_width = chart_specs.get('width')  # Maximum width of the chart

        self._x_elements_coordinates: Series | None = None
        self._x_data: Series | None = None
        self._x_encoding: str = ''
        self._x_offset: float = 0

        self._y_elements_coordinates: Series | None = None
        self._y_data: Series | None = None
        self._y_encoding: str = ''
        self._y_offset: float = 0

        self._z_elements_coordinates: Series | None = None
        self._z_data: Series | None = None
        self._z_encoding: str = ''
        self._z_offset: float = 0

        self._process_channels('color', 'x', 'y', 'z')  # Process and set self._{axis} attributes

    def _apply_axis_offset(self, coordinates: Series, axis: str, invert: bool = False,
                           extra_offset: float = 0) -> Series:
        min_val = coordinates.min()
        offset = abs(min_val) + extra_offset if min_val < 0 else 0

        if invert:
            result = -(offset + coordinates)
            setattr(self, f'_{axis}_offset', -offset)
        else:
            result = offset + coordinates
            setattr(self, f'_{axis}_offset', offset)

        setattr(self, f'_{axis}_elements_coordinates', result)
        return result

    def _correct_axes_position(self, elem_size: float | None) -> None:
        """
        Corrects the axes' position for inner the calculations and processing.
        Must be called by child classes when initiating.
        """

        def _calculate_axis_size(axis_data: Series, default_axis_size: float) -> float:
            if elem_size is None or axis_data is None:  # User did not define bars' size, or there is no data
                return default_axis_size  # Set default value

            if _translate_dtype_into_encoding(axis_data.dtype) == 'quantitative':
                return default_axis_size  # User did not define bars' size or axis is quantitative

            return elem_size * axis_data.n_unique()

        # X-axis
        if self._chart_width is None:  # User did not define chart width
            self._chart_width = _calculate_axis_size(self._x_data, DEFAULT_CHART_WIDTH)

        # Y-axis
        if self._chart_height is None:  # User did not define chart height
            self._chart_height = _calculate_axis_size(self._y_data, DEFAULT_CHART_HEIGHT)

        # Z-axis
        if self._chart_depth is None:  # User did not define chart depth
            self._chart_depth = _calculate_axis_size(self._z_data, DEFAULT_CHART_DEPTH)

        self._relative_bottom_left_corner_position = (
            f'{-self._chart_width / 2} '
            f'{-self._chart_height / 2} '
            f'{self._chart_depth / 2}'
        )

    def get_axes_specs(self) -> dict:
        """Returns a dictionary with the specifications for each axis of the chart."""
        if self._raw_data.is_empty():  # There is no data to display
            return {}

        axis_specs = {}

        for axis in ('x', 'y', 'z'):  # type: Literal['x', 'y', 'z']
            encoding = self._encoding.get(axis)
            if not encoding:
                continue

            if not encoding.get('axis', True):
                continue  # Display axis if key 'axis' not found (default display axis) or True

            axis_specs[axis] = AxisCreator.create_axis_specs(
                axis=axis,
                axis_data=getattr(self, f'_{axis}_data'),
                axis_encoding=getattr(self, f'_{axis}_encoding'),
                axis_size=getattr(self, f'_chart_{self._AXIS_SIZE_MAP[axis]}'),
                elements_coords=getattr(self, f'_{axis}_elements_coordinates'),
                x_offset=0 if axis == 'x' else self._x_offset,
                y_offset=0 if axis == 'y' else self._y_offset,
                z_offset=0 if axis == 'z' else self._z_offset,
            )

        return axis_specs

    def get_legend_elements(self, filtered_by_params: bool = False) -> list[ElementCreator]:
        """Returns a list for the elements of the legend."""
        if self._color_data is None:
            return []

        color_mapping = {}
        colors = self._set_elements_colors()
        for m, c in zip(self._color_data, colors):
            color_mapping.setdefault(m, c)

        center_x_pos = self._chart_width + LEGEND_WIDTH - 1

        # Plane
        plane_height = LEGEND_HEIGHT_PER_ELEMENT * len(color_mapping)
        plane = PlaneCreator({
            'position': f'{center_x_pos} {self._chart_height / 2} 0',
            'height': plane_height, 'width': LEGEND_WIDTH
        }, filtered_by_params=filtered_by_params)

        # Text
        text_base_y = self._chart_height / 2 - plane_height / 2 + LEGEND_HEIGHT_PER_ELEMENT / 2
        text = [
            TextCreator({
                'position': f'{center_x_pos} {text_base_y + LEGEND_HEIGHT_PER_ELEMENT * index} 0',
                'value': item[0].title(), 'color': item[1],
                'align': 'center', 'scale': LABELS_SCALE
            }, filtered_by_params=filtered_by_params)
            for index, item in enumerate(color_mapping.items())
        ]

        return [plane, *text]

    @staticmethod
    def set_elems_coordinates_for_quantitative_axis(axis_data: Series, axis_size: float,
                                                    extremes_offset: float) -> Series:
        """
        Returns a Series with the positions for each element in the quantitative axis.

        Parameters
        ----------
        axis_data: Series
            The data of the quantitative axis.
        axis_size : float
            The total size of the axis.
        extremes_offset : float
            The offset used in each extreme of the axis, so the elements do not exceed the chart dimensions.
        """
        if axis_data.dtype == pl.String:
            axis_data = axis_data.cast(pl.Categorical).to_physical()

        max_value, min_value = axis_data.max(), axis_data.min()  # For proportions
        range_value = max_value - min_value  # Range (positive value)
        if range_value == 0:  # All the values are the same
            return pl.repeat(
                value=axis_size / 2,  # Center elements in the axis
                n=axis_data.len(),
                eager=True  # Returns a Series
            )

        usable_axis_size = axis_size - (2 * extremes_offset)  # Reduce the axis space size
        if max_value < 0:  # All data is negative
            scale_factor = usable_axis_size / -min_value
            final_offset = -extremes_offset  # Negative offset
        elif min_value >= 0:  # All data is positive (including 0)
            scale_factor = usable_axis_size / max_value
            final_offset = extremes_offset  # Positive offset
        else:  # Positive and negative data
            scale_factor = usable_axis_size / range_value
            final_offset = 0
        return axis_data * scale_factor + final_offset  # Add final offset to center

    @staticmethod
    def set_elems_coordinates_for_nominal_axis(axis_data: Series, axis_size: float, extremes_offset: float) -> Series:
        """
        Returns a Series with the positions for each element in the nominal axis.

        Parameters
        ----------
        axis_data : Series
            The data of the nominal axis.
        axis_size : float
            The total size of the axis.
        extremes_offset : float
            The offset used in each extreme of the axis, so the elements do not exceed the chart dimensions.
        """
        category_codes = axis_data.cast(pl.Categorical).to_physical()
        unique_categories = axis_data.n_unique()

        step = (axis_size - 2 * extremes_offset) / (unique_categories - 1) if unique_categories > 1 else 0
        return (extremes_offset + step * category_codes).cast(pl.Float32)


class NonAxisChannelChartCreator(ChartCreator):
    """Chart creator base class for charts that have channels but do not have XYZ axis."""

    def get_axes_specs(self):
        """Returns a Series with the specifications for each axis of the chart."""
        return {}  # Returns an empty dictionary, because it has no axis


# Second-level subclasses of ChartCreator.
class ArcChartCreator(NonAxisChannelChartCreator):
    """Arc chart creator class."""

    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._radius = chart_specs['mark'].get('radius', DEFAULT_PIE_RADIUS) \
            if isinstance(chart_specs['mark'], dict) else DEFAULT_PIE_RADIUS
        self._set_rotation()

        self._theta_data: Series | None = None
        self._theta_encoding: str = ''

        self._process_channels('color', 'theta')

    def _set_rotation(self):
        """Sets the rotation of the pie chart."""
        pie_rotation = DEFAULT_PIE_ROTATION.split()  # Default rotation for the pie chart to look at the camera
        self._x_rotation = self._x_rotation + float(pie_rotation[0])
        self._y_rotation = self._y_rotation + float(pie_rotation[1])
        self._z_rotation = self._z_rotation + float(pie_rotation[2])

    def _set_elements_theta(self) -> tuple[Series, Series]:
        """Returns a tuple with a Series storing the theta start of each element, and another storing theta length."""
        abs_theta_data = self._theta_data.abs()
        sum_data = abs_theta_data.sum()  # Sum all the values
        theta_length = (360 / sum_data) * abs_theta_data  # Series of theta lengths (in degrees)
        theta_start = theta_length.cum_sum().shift(1).fill_null(0)  # Accumulative sum (first value is 0)
        return theta_start.alias('theta_start'), theta_length.alias('theta_length')

    def get_elements(self, filtered_by_params: bool) -> list[ElementCreator]:
        """Returns a list of each element composing the chart."""
        if self._raw_data.is_empty():  # There is no data to display
            return []

        data_length = self._raw_data.height  # Number of rows in data

        # Axis
        zeros = pl.repeat(0, data_length, eager=True)
        x_coordinates = y_coordinates = z_coordinates = zeros

        # Color and theta
        if self._theta_encoding != 'quantitative':
            raise ValueError('Theta-channel data must be quantitative')

        colors = self._set_elements_colors()
        theta_starts, theta_lengths = self._set_elements_theta()

        # Depth
        depth = pl.repeat(
            value=self._chart_depth,
            n=data_length,
            eager=True  # Returns a Series
        ).alias('depth')

        # Radius
        radius = pl.repeat(
            value=self._radius,
            n=data_length,
            eager=True  # Returns a Series
        ).alias('radius')

        # Information display
        info = pl.select(pl.concat_str(
            [self._color_data.cast(pl.String), self._theta_data.cast(pl.String)],
            separator=' : ',
        ).fill_null('?').alias('id')).to_series()

        # Return values
        temp_dict = {
            'info': info,
            'height': depth,  # Using height, as pie's slices are rotated cylinders
            'position': pl.select(pl.concat_str(
                [x_coordinates, y_coordinates, z_coordinates],
                separator=' '
            ).alias('position')).to_series(),
            'radius': radius,
            'theta_start': theta_starts,
            'theta_length': theta_lengths,
            'color': colors
        }

        # Selection
        self._add_selection_to_specs(temp_dict)

        elements_specs = pl.from_dict(temp_dict).to_dicts()  # Transform into a list of dictionaries
        return [
            CylinderCreator(specification, filtered_by_params=filtered_by_params)
            for specification in elements_specs
        ]


class BarChartCreator(XYZAxisChannelChartCreator):
    """Bar chart creator class."""

    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._bar_size_if_nominal_axis: float | None = chart_specs['mark'].get('size') \
            if isinstance(chart_specs['mark'], dict) else None
        self._correct_axes_position(elem_size=self._bar_size_if_nominal_axis)

    def _set_bars_coords_size_in_axis(self, axis_data: Series, axis_name: Literal['x', 'y', 'z'],
                                      encoding_type: str) -> tuple[Series, Series]:
        """
        Returns a tuple of Series.
        The first contains the axis coordinates of each bar for the given axis.
        The second contains the dimensions of each bar for the given axis.
        """
        try:
            axis_size = getattr(self, f'_chart_{self._AXIS_SIZE_MAP[axis_name]}')  # Get axis dimension
            bars_size_alias = self._AXIS_SIZE_MAP[axis_name]  # Get alias of bar size Series depending on axis
        except KeyError:  # pragma: no cover (should never enter here, except code errors)
            raise RuntimeError('Unreachable code. Axis must be x or y or z')

        if axis_data is None:
            coordinates = pl.repeat(
                value=axis_size / 2,
                n=self._raw_data.height,  # Number of rows in data
                eager=True  # Returns a Series
            )
            bars_axis_size = 2 * coordinates  # Multiplied by 2 because of how boxes are created
        else:
            if encoding_type == 'quantitative':
                coordinates = 0.5 * self.set_elems_coordinates_for_quantitative_axis(  # Half because of bar's creation
                    axis_data=axis_data,
                    axis_size=axis_size,
                    extremes_offset=0  # The greatest bar reaches axis size
                )
                bars_axis_size = 2 * coordinates.abs()
            elif encoding_type == 'nominal':
                unique_values = axis_data.n_unique()
                if self._bar_size_if_nominal_axis is not None:  # User defined bars' size
                    if self._bar_size_if_nominal_axis * unique_values > axis_size:  # Bars would overlap
                        bar_size = axis_size / unique_values  # Adjust bars' axis size automatically
                        warnings.warn(
                            f'Defined bar size will make bars overlap on the {axis_name}-axis, adjusting automatically '
                            f'for this axis. Consider changing {bars_size_alias}.'
                        )

                    else:  # Bars do not overlap with user's defined size
                        bar_size = self._bar_size_if_nominal_axis  # Use user's defined size
                else:  # User did not define bars' size
                    bar_size = axis_size / unique_values * (1 - DEFAULT_BAR_PADDING)  # Adjust bars' axis size

                coordinates = self.set_elems_coordinates_for_nominal_axis(
                    axis_data=axis_data, axis_size=axis_size,
                    extremes_offset=bar_size / 2
                )
                bars_axis_size = pl.repeat(
                    value=bar_size,
                    n=axis_data.len(),
                    eager=True  # Returns a Series
                )
            else:
                raise ValueError(f'Invalid encoding type: {encoding_type}.')
        return coordinates.alias(f'{axis_name}_coordinates'), bars_axis_size.alias(bars_size_alias)

    def get_elements(self, filtered_by_params: bool) -> list[ElementCreator]:
        """Returns a list of each element composing the chart."""
        if self._raw_data.is_empty():  # There is no data to display
            return []

        # XYZ-axis
        x_coordinates, bar_widths = self._set_bars_coords_size_in_axis(
            axis_data=self._x_data, axis_name='x', encoding_type=self._x_encoding
        )
        self._apply_axis_offset(x_coordinates, 'x')

        y_coordinates, bar_heights = self._set_bars_coords_size_in_axis(
            axis_data=self._y_data, axis_name='y', encoding_type=self._y_encoding
        )
        self._apply_axis_offset(y_coordinates, 'y')

        z_coordinates, bar_depths = self._set_bars_coords_size_in_axis(
            axis_data=self._z_data, axis_name='z', encoding_type=self._z_encoding
        )
        self._apply_axis_offset(z_coordinates, 'z', invert=True)  # Invert sign (to go deep)

        # Color
        colors = self._set_elements_colors()

        # Information display
        info_series = [
            s.cast(pl.String)
            for s in (self._x_data, self._y_data, self._z_data)
            if s is not None
        ]

        info = pl.select(pl.concat_str(
            info_series,
            separator=' : '
        ).fill_null('?').alias('id')).to_series()

        # Return values
        temp_dict = {
            'info': info,
            'position': pl.select(pl.concat_str(
                [self._x_elements_coordinates, self._y_elements_coordinates, self._z_elements_coordinates],
                separator=' '
            ).alias('position')).to_series(),
            'width': bar_widths,
            'height': bar_heights,
            'depth': bar_depths,
            'color': colors
        }

        # Selection
        self._add_selection_to_specs(temp_dict)

        elements_specs = pl.from_dict(temp_dict).to_dicts()  # Transform into a list of dictionaries
        return [
            BoxCreator(specification, filtered_by_params=filtered_by_params)
            for specification in elements_specs
        ]


class LineChartCreator(XYZAxisChannelChartCreator):
    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._correct_axes_position(elem_size=DEFAULT_VERTICES_SPACING)
        self._line_color = chart_specs['mark'].get('color', DEFAULT_ELEMENTS_COLOR_IN_CHART) \
            if isinstance(chart_specs['mark'], dict) else DEFAULT_ELEMENTS_COLOR_IN_CHART
        self._display_points_in_vertices = chart_specs['mark'].get('point', DEFAULT_VERTICES_POINT_DISPLAY) \
            if isinstance(chart_specs['mark'], dict) else DEFAULT_VERTICES_POINT_DISPLAY
        self._marker_bbox_size_half = _calculate_point_radius(DEFAULT_VERTICES_POINT_VOLUME) \
            if self._display_points_in_vertices else 0  # Half of the markers bounding box's axes size

    def _set_extremes_coords_in_axis(self, axis_data: Series, axis_name: Literal['x', 'y', 'z'],
                                     encoding_type: str) -> Series:
        """Returns a Series containing the coordinates for each extreme of the line, for the given axis."""
        attr_name = self._AXIS_SIZE_MAP.get(axis_name)
        if not attr_name:  # pragma: no cover (this method is internally called)
            raise RuntimeError('Unreachable code. Parameter axis_name is not correct')

        axis_size = getattr(self, f'_chart_{attr_name}')  # Get axis dimensions depending on the given axis

        if axis_data is None:
            coordinates = pl.repeat(
                value=self._marker_bbox_size_half,
                n=self._raw_data.height,  # Number of rows in data
                eager=True  # Returns a Series
            )
        else:
            if encoding_type == 'quantitative':
                coordinates = self.set_elems_coordinates_for_quantitative_axis(
                    axis_data=axis_data,
                    axis_size=axis_size,
                    extremes_offset=self._marker_bbox_size_half
                )
            elif encoding_type == 'nominal':
                coordinates = self.set_elems_coordinates_for_nominal_axis(
                    axis_data=axis_data, axis_size=axis_size,
                    extremes_offset=self._marker_bbox_size_half
                )
            else:
                raise ValueError(f'Invalid encoding type: {encoding_type}.')
        return coordinates.alias(f'{axis_name}_coordinates')

    def get_elements(self, filtered_by_params: bool) -> list[ElementCreator]:
        """Returns a list of each element composing the chart."""
        if self._raw_data.is_empty():  # There is no data to display
            return []

        x_coordinates = self._set_extremes_coords_in_axis(self._x_data, axis_name='x', encoding_type=self._x_encoding)
        self._apply_axis_offset(x_coordinates, 'x')

        y_coordinates = self._set_extremes_coords_in_axis(self._y_data, axis_name='y', encoding_type=self._y_encoding)
        self._apply_axis_offset(y_coordinates, 'y')

        z_coordinates = self._set_extremes_coords_in_axis(self._z_data, axis_name='z', encoding_type=self._z_encoding)
        self._apply_axis_offset(z_coordinates, 'z', invert=True)  # Invert sign (to go deep)

        # Colors
        colors = self._set_elements_colors()

        # Positions
        positions = pl.select(pl.concat_str(
            [self._x_elements_coordinates, self._y_elements_coordinates, self._z_elements_coordinates],
            separator=' '
        ).alias('position')).to_series()

        # Lines
        lines_df = (
            pl.DataFrame({'start': positions, 'color': colors})
            .with_columns(pl.col('start').shift(-1).over('color').alias('end'))  # Shift one position up (of same color)
            .drop_nulls('end')  # Remove last row (NULL value)
        )

        # Points
        info_series = [
            s.cast(pl.String)
            for s in (self._x_data, self._y_data, self._z_data)
            if s is not None
        ]
        info = pl.select(pl.concat_str(info_series, separator=' : ').fill_null('?').alias('id')).to_series()

        points_df = pl.DataFrame({
            'position': positions,
            'info': info,
            'color': colors,
            'radius': pl.repeat(
                _calculate_point_radius(DEFAULT_VERTICES_POINT_VOLUME),
                n=self._raw_data.height,
                eager=True
            )
        })

        # Selection
        for df in (lines_df, points_df):
            self._add_selection_to_specs(df.to_dict(as_series=False))

        # Return elements
        elements_lines = [LineCreator(spec, filtered_by_params=filtered_by_params) for spec in lines_df.to_dicts()]
        elements_points = (
            [SphereCreator(spec, filtered_by_params=filtered_by_params) for spec in points_df.to_dicts()]
            if self._display_points_in_vertices
            else []
        )

        return elements_lines + elements_points

class PointChartCreator(XYZAxisChannelChartCreator):
    """Point chart creator class."""

    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        max_sphere_volume: float = chart_specs['mark'].get('size', DEFAULT_POINT_VOLUME) \
            if isinstance(chart_specs['mark'], dict) else DEFAULT_POINT_VOLUME
        self._max_radius = _calculate_point_radius(max_sphere_volume)
        self._correct_axes_position(elem_size=self._max_radius * 2)

        self._size_data: Series | None = None
        self._size_encoding: str = ''

        self._process_channels('color', 'size')  # Process and set self._{ch} attributes

    def _set_points_coords_in_axis(self, axis_data: Series, axis_name: Literal['x', 'y', 'z'],
                                   encoding_type: str) -> Series:
        """Returns a Series containing the coordinates for each point of the chart, for the given axis."""
        attr_name = self._AXIS_SIZE_MAP.get(axis_name)
        if not attr_name:  # pragma: no cover (this method is internally called)
            raise RuntimeError('Unreachable code. Parameter axis_name is not correct')

        axis_size = getattr(self, f'_chart_{attr_name}')  # Get axis dimensions depending on the given axis

        if axis_data is None:
            coordinates = pl.repeat(
                value=axis_size / 2,  # Center points in the axis
                n=self._raw_data.height,  # Number of rows in data
                eager=True  # Returns a Series
            )
        else:
            if encoding_type == 'quantitative':
                coordinates = self.set_elems_coordinates_for_quantitative_axis(
                    axis_data=axis_data,
                    axis_size=axis_size,
                    extremes_offset=self._max_radius + EPSILON  # Points do not exceed the dimensions of the axis
                )
            elif encoding_type == 'nominal':
                coordinates = self.set_elems_coordinates_for_nominal_axis(
                    axis_data=axis_data, axis_size=axis_size,
                    extremes_offset=self._max_radius + EPSILON  # Points do not exceed the dimensions of the axis
                )
            else:
                raise ValueError(f'Invalid encoding type: {encoding_type}.')
        return coordinates.alias(f'{axis_name}_coordinates')

    def _set_points_radius(self) -> Series:
        """Returns a Series of the radius for each point composing the bubble chart."""
        if self._size_encoding and self._size_encoding != 'quantitative':
            raise ValueError(ERROR_MESSAGES['SIZE_ENCODING_NOT_QUANTITATIVE'].format(size_encoding=self._size_encoding))

        if self._size_data is None:  # Scatter plot (same radius for all points)
            points_radius = pl.repeat(
                value=self._max_radius,  # Same radius for all points
                n=self._raw_data.height,  # Number of rows in data
                eager=True  # Returns a Series
            )
        else:  # Bubbles plot (the size of the point depends on the value of the field)
            max_value = self._size_data.max()
            points_radius = (self._size_data / max_value) * self._max_radius
        return points_radius.alias('radius')

    def get_elements(self, filtered_by_params: bool) -> list[ElementCreator]:
        """Returns a list of each element composing the chart."""
        if self._raw_data.is_empty():  # There is no data to display
            return []

        # Channels
        colors = self._set_elements_colors()
        radius = self._set_points_radius()

        x_coordinates = self._set_points_coords_in_axis(
            axis_data=self._x_data, axis_name='x', encoding_type=self._x_encoding
        )
        self._apply_axis_offset(x_coordinates, 'x', extra_offset=self._max_radius)

        y_coordinates = self._set_points_coords_in_axis(
            axis_data=self._y_data, axis_name='y', encoding_type=self._y_encoding
        )
        self._apply_axis_offset(y_coordinates, 'y', extra_offset=self._max_radius)

        z_coordinates = self._set_points_coords_in_axis(
            axis_data=self._z_data, axis_name='z', encoding_type=self._z_encoding
        )
        self._apply_axis_offset(z_coordinates, 'z', invert=True, extra_offset=self._max_radius)  # Invert (go deep)

        # Information display
        info_series = [
            s.cast(pl.String)
            for s in (self._x_data, self._y_data, self._z_data)
            if s is not None
        ]

        info = pl.select(pl.concat_str(
            info_series,
            separator=' : ',
        ).fill_null('?').alias('id')).to_series()

        # Return values
        temp_dict = {
            'info': info,
            'position': pl.select(pl.concat_str(
                [self._x_elements_coordinates, self._y_elements_coordinates, self._z_elements_coordinates],
                separator=' '
            ).alias('position')).to_series(),
            'radius': radius,
            'color': colors,
        }

        # Selection
        self._add_selection_to_specs(temp_dict)

        elements_specs = pl.from_dict(temp_dict).to_dicts()  # Transform into a list of dictionaries
        return [
            SphereCreator(specifications, filtered_by_params=filtered_by_params)
            for specifications in elements_specs
        ]


# Add classes to CREATOR_MAP
CREATOR_MAP.update({
    'arc': ArcChartCreator,
    'bar': BarChartCreator,
    'line': LineChartCreator,
    'point': PointChartCreator,
})



"""AframeXR axis creator"""

import polars as pl

from polars import Series
from typing import Literal

from .constants import *


_X_AXIS_LABELS_ROTATION = '-90 0 -90'
_Y_AXIS_LABELS_ROTATION = '0 0 0'
_Z_AXIS_LABELS_ROTATION = '-90 0 0'


def _get_labels_coords_for_quantitative_axis(axis_data: Series, axis_size: float) -> Series:
    """Returns the coordinates for the labels of the quantitative axis."""
    unique_values = axis_data.n_unique()

    if unique_values == 1:  # All the values are the same
        return Series([axis_size / 2])  # Only one tick is placed in the axis

    num_samples = unique_values if axis_data.dtype == pl.String else DEFAULT_NUM_OF_TICKS_IF_QUANTITATIVE_AXIS
    return pl.linear_space(  # Equally spaced values
            start=START_LABEL_OFFSET,  # Offset for the lowest label (for not being on the ground)
            end=axis_size,
            num_samples=num_samples,
            eager=True  # Returns a Series
        )

def _get_labels_values_for_quantitative_axis(axis_data: Series) -> Series:
    """Returns the values for the labels of the quantitative axis."""
    if axis_data.dtype == pl.String:  # Axis data contains nominal values, but user wants to encode as quantitative
        return axis_data.unique(maintain_order=True)  # Return the same values

    min_value, max_value = axis_data.min(), axis_data.max()

    if max_value == min_value:  # All the values are the same
        labels_values = Series([axis_data[0]])  # Only one tick is placed in the axis
    else:
        if max_value < 0:  # All data is negative
            start = min_value
            end = 0
        elif min_value > 0:  # All data is positive
            start = 0
            end = max_value
        else:
            start, end = min_value, max_value

        labels_values = pl.linear_space(
            start=start,
            end=end,
            num_samples=DEFAULT_NUM_OF_TICKS_IF_QUANTITATIVE_AXIS,
            eager=True  # Returns a Series
        )
    return labels_values


class AxisCreator:
    """Axis creator class."""

    @staticmethod
    def create_axis_html(start: str, end: str) -> str:
        """
        Create a line for the axis and returns its HTML.

        Parameters
        ----------
        start : str
            The base position of each axis.
        end : str
            The end position of the axis.
        """
        return f'<a-entity line="start: {start}; end: {end}; color: black"></a-entity>'

    @staticmethod
    def create_axis_specs(axis: Literal['x', 'y', 'z'], axis_data: Series, axis_encoding: str, axis_size: float,
                          elements_coords: Series, x_offset: float, y_offset: float, z_offset: float) -> dict:
        """Returns the axis specifications for x, y or z axis depending on its encoding."""
        axis_specs = {'start': None, 'end': None, 'labels_pos': [], 'labels_values': [], 'labels_rotation': '',
                      'labels_align': None}

        if axis_encoding == 'quantitative':
            coords = _get_labels_coords_for_quantitative_axis(axis_data, axis_size)
            labels_values = _get_labels_values_for_quantitative_axis(axis_data)
        elif axis_encoding == 'nominal':
            coords = elements_coords.unique(maintain_order=True)  # Align labels with elements
            labels_values = axis_data.unique(maintain_order=True)
        else:  # pragma: no cover (Encoding type must have been checked before)
            raise RuntimeError(f'Unreachable code. Check encoding type: {axis_encoding}')

        axis_specs['start'] = f'{x_offset} {y_offset} {z_offset}'
        if axis == 'x':
            axis_specs['end'] = f'{axis_size} {y_offset} {z_offset}'
            axis_specs['labels_pos'] = (coords.cast(pl.String) + f' {LABELS_Y_DELTA} {X_LABELS_Z_DELTA}').to_list()
            axis_specs['labels_rotation'] = _X_AXIS_LABELS_ROTATION
            axis_specs['labels_align'] = 'left'
        elif axis == 'y':
            axis_specs['end'] = f'{x_offset} {axis_size} {z_offset}'
            axis_specs['labels_pos'] = (f'{LABELS_X_DELTA} ' + coords.cast(pl.String) + ' 0').to_list()
            axis_specs['labels_rotation'] = _Y_AXIS_LABELS_ROTATION
            axis_specs['labels_align'] = 'right'
        elif axis == 'z':
            axis_specs['end'] = f'{x_offset} {y_offset} {-axis_size}'  # Negative axis size to go deep
            axis_specs['labels_pos'] = (f'{LABELS_X_DELTA} {LABELS_Y_DELTA} ' + coords.cast(pl.String)).to_list()
            axis_specs['labels_rotation'] = _Z_AXIS_LABELS_ROTATION
            axis_specs['labels_align'] = 'right'
        else:  # pragma: no cover (this method is only called by inner code methods; should be OK)
            raise RuntimeError('Unreachable code. Axis must be x or y or z')

        axis_specs['labels_values'] = labels_values.to_list()
        return axis_specs



"""AframeXR utils"""

from .axis_creator import *
from .chart_creator import *
from .constants import *
from .entities_html_creator import *
from .chart_creator import *
from .scene_creator import *
from .validators import *
